from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Dict
from typing import Protocol
from typing import Type

from geneticengine.core.grammar import Grammar
from geneticengine.core.random.sources import Source


class MetaHandlerGenerator(Protocol):
    """MetaHandlers are type refinements. They override the generation procedure of the base type."""

    def generate(
        self,
        r: Source,
        g: Grammar,
        rec,
        new_symbol,
        depth: int,
        base_type,
        context: dict[str, str],
    ):
        """
        Generates an instance of type base_type, according to some criterion.

        :param Source r: Random source for generation
        :param Grammar g: Grammar to follow in the generation
        :param Callable[[int, Type], Any] rec: The method to generate a new instance of type and maximum depth d
        :param int depth: the current depth budget
        :param Type base_type: The inner type being annotated
        :param str argname: The name of the field of the parent object which is being generated
        :param Dict[str, Type] context: The names and types of all fields in the parent object
        """
        ...


def is_metahandler(ty: type) -> bool:
    """
    Returns if type is a metahandler.
    AnnotatedType[int, IntRange(3,10)] is an example of a Metahandler.

    Verification is done using the __metadata__, which is the first argument of Annotated
    """
    return hasattr(ty, "__metadata__")



from __future__ import annotations

from dataclasses import dataclass
from typing import Any
from typing import Callable

from geneticengine.grammars.coding.classes import Condition
from geneticengine.grammars.coding.classes import Number


@dataclass
class Equals(Condition):
    left: Number
    right: Number

    def evaluate(self, **kwargs) -> bool:
        return self.left.evaluate(**kwargs) == self.right.evaluate(**kwargs)

    def evaluate_lines(self, **kwargs) -> Callable[[Any], bool]:
        return lambda line: self.left.evaluate_lines(**kwargs)(
            line,
        ) == self.right.evaluate_lines(**kwargs)(line)

    def __str__(self):
        return f"({self.left} == {self.right})"


@dataclass
class NotEquals(Condition):
    left: Number
    right: Number

    def evaluate(self, **kwargs) -> bool:
        return self.left.evaluate(**kwargs) != self.right.evaluate(**kwargs)

    def evaluate_lines(self, **kwargs) -> Callable[[Any], bool]:
        return lambda line: self.left.evaluate_lines(**kwargs)(
            line,
        ) != self.right.evaluate_lines(**kwargs)(line)

    def __str__(self):
        return f"({self.left} != {self.right})"


@dataclass
class GreaterThan(Condition):
    left: Number
    right: Number

    def evaluate(self, **kwargs) -> bool:
        return self.left.evaluate(**kwargs) > self.right.evaluate(**kwargs)

    def evaluate_lines(self, **kwargs) -> Callable[[Any], bool]:
        return lambda line: self.left.evaluate_lines(**kwargs)(
            line,
        ) > self.right.evaluate_lines(**kwargs)(line)

    def __str__(self):
        return f"({self.left} > {self.right})"


@dataclass
class GreaterOrEqualThan(Condition):
    left: Number
    right: Number

    def evaluate(self, **kwargs) -> bool:
        return self.left.evaluate(**kwargs) >= self.right.evaluate(**kwargs)

    def evaluate_lines(self, **kwargs) -> Callable[[Any], bool]:
        return lambda line: self.left.evaluate_lines(**kwargs)(
            line,
        ) >= self.right.evaluate_lines(**kwargs)(line)

    def __str__(self):
        return f"({self.left} >= {self.right})"


@dataclass
class LessThan(Condition):
    left: Number
    right: Number

    def evaluate(self, **kwargs) -> bool:
        return self.left.evaluate(**kwargs) < self.right.evaluate(**kwargs)

    def evaluate_lines(self, **kwargs) -> Callable[[Any], bool]:
        return lambda line: self.left.evaluate_lines(**kwargs)(
            line,
        ) < self.right.evaluate_lines(**kwargs)(line)

    def __str__(self):
        return f"({self.left} < {self.right})"


@dataclass
class LessOrEqualThan(Condition):
    left: Number
    right: Number

    def evaluate(self, **kwargs) -> bool:
        return self.left.evaluate(**kwargs) <= self.right.evaluate(**kwargs)

    def evaluate_lines(self, **kwargs) -> Callable[[Any], bool]:
        return lambda line: self.left.evaluate_lines(**kwargs)(
            line,
        ) <= self.right.evaluate_lines(**kwargs)(line)

    def __str__(self):
        return f"({self.left} <= {self.right})"


@dataclass
class Is(Condition):
    left: Number
    right: Number

    def evaluate(self, **kwargs) -> bool:
        return self.left.evaluate(**kwargs) is self.right.evaluate(**kwargs)

    def evaluate_lines(self, **kwargs) -> Callable[[Any], bool]:
        return lambda line: self.left.evaluate_lines(**kwargs)(
            line,
        ) is self.right.evaluate_lines(**kwargs)(line)

    def __str__(self):
        return f"({self.left} is {self.right})"


@dataclass
class IsNot(Condition):
    left: Number
    right: Number

    def evaluate(self, **kwargs) -> bool:
        return self.left.evaluate(**kwargs) is not self.right.evaluate(**kwargs)

    def evaluate_lines(self, **kwargs) -> Callable[[Any], bool]:
        return lambda line: self.left.evaluate_lines(**kwargs)(
            line,
        ) is not self.right.evaluate_lines(**kwargs)(line)

    def __str__(self):
        return f"({self.left} is not {self.right})"


all_operators = [
    Equals,
    NotEquals,
    GreaterOrEqualThan,
    GreaterThan,
    LessOrEqualThan,
    LessThan,
    Is,
    IsNot,
]



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Callable
from typing import Optional
from typing import TypeVar
from typing import Union


FitnessType = Union[float, list[float]]
P = TypeVar("P")


class Problem(ABC):
    """
    An Abstract class that SingleObjectiveProblem and MultiObjectiveProblem extends to

    Args:
        minimize (bool | list[bool]): When switch on, the fitness function is reversed, so that a higher result from the fitness function corresponds to a less fit solution.
        fitness_function (Callable[[P], float] | Callable[[P], list[float]]): The fitness function. Should take in any valid individual and return a float or a list of floats, depending if its a single objetive problem or a multi objective problem.
    """

    minimize: bool | list[bool]
    fitness_function: Callable[[P], float] | Callable[[P], list[float]]

    @abstractmethod 
    def evaluate(self, p: P) -> FitnessType:
        ...

    def solved(self, best_fitness: FitnessType):
        return False


@dataclass
class SingleObjectiveProblem(Problem):
    """
    SingleObjectiveProblem is a class that extends the Problem class

    Args:
        minimize (bool): When switch on, the fitness function is reversed, so that a higher result from the fitness function corresponds to a less fit solution.
        fitness_function (Callable[[P], float]): The fitness function. Should take in any valid individual and return a float.
        target_fitness (Optional[float]): Sets a target fitness. When this fitness is reached, the algorithm stops running (default = None).
    """

    minimize: bool
    fitness_function: Callable[[P], float]
    target_fitness: float | None

    def evaluate(self, p: P) -> float:
        return float(self.fitness_function(p))

    def solved(self, best_fitness: FitnessType):
        assert isinstance(best_fitness, float)
        if not self.target_fitness:
            return False
        elif self.minimize:
            return best_fitness <= self.target_fitness
        else:
            return best_fitness >= self.target_fitness


@dataclass
class MultiObjectiveProblem(Problem):
    """
    MultiObjectiveProblem is a class that extends the Problem class

    Args:
        minimize (list[bool]): When switch on, the fitness function is reversed, so that a higher result from the fitness function corresponds to a less fit solution.
        fitness_function (Callable[[P], list[bool]]): The fitness function. Should take in any valid individual and return a list of float.
        best_individual_criteria_function (Optional(Callable[[P], float]): This function allow the user to choose how to find the best individual in a generation (default = None , this means that the individual with the best fitness is the one considered as the best in that generation)
    """

    minimize: list[bool]
    fitness_function: Callable[[P], list[float]]
    best_individual_criteria_function: Callable[[P], float] | None = None

    def number_of_objectives(self):
        return len(self.minimize)

    def evaluate(self, p: P) -> list[float]:
        return [float(x) for x in self.fitness_function(p)]


def wrap_depth_minimization(p: SingleObjectiveProblem) -> SingleObjectiveProblem:
    """
    This wrapper takes a SingleObjectiveProblem and adds a penalty for bigger trees.
    """

    def w(i):
        if p.minimize:
            return p.fitness_function(i) + i.gengy_nodes * 10**-25
        else:
            return p.fitness_function(i) - i.gengy_nodes * 10**-25

    return SingleObjectiveProblem(
        minimize=p.minimize,
        fitness_function=w,
        target_fitness=None,
    )


def process_problem(
    problem: Problem | None,
    evaluation_function: Callable[[P], float] | None = None,  # DEPRECATE in the next version
    minimize: bool = False,  # DEPRECATE in the next version
    target_fitness: float | None = None,  # DEPRECATE in the next version
) -> Problem:
    """
    This function is a placeholder until we deprecate all the old usage of GP class.
    """
    if problem:
        return problem
    elif isinstance(minimize, list) and evaluation_function:
        return MultiObjectiveProblem(minimize, evaluation_function)
    elif isinstance(minimize, bool) and evaluation_function:
        return SingleObjectiveProblem(minimize, evaluation_function, target_fitness)
    else:
        raise NotImplementedError(
            "This combination of parameters to define the problem is not valid",
        )


def wrap_depth(p: Problem, favor_less_complex_trees: bool = False):
    if isinstance(p, SingleObjectiveProblem):
        if favor_less_complex_trees:
            return wrap_depth_minimization(p)
        else:
            return p
    else:
        assert isinstance(p, MultiObjectiveProblem)
        return p



from __future__ import annotations

import glob
from typing import Any
from typing import List

import matplotlib  # type:ignore
import matplotlib.pyplot as plt  # type:ignore
import numpy as np
import pandas as pd
import seaborn as sns # type: ignore

from geneticengine.exceptions import GeneticEngineError
from geneticengine.visualization.utils import *

matplotlib.rcParams["pdf.fonttype"] = 42
matplotlib.rcParams["ps.fonttype"] = 42

def plot_comparison(folder_names: list, labels: list, labels_name: str = 'Labels', x_axis: str = 'Generations', y_axis: str = 'Fitness', title: str = 'Fitness comparison', file_name = None):
    '''
        Plots a figure with lines for each folder (average with shades for std) with each folder from folder_names named with the corresponding labels_name.
    '''
    assert len(folder_names) == len(labels)
    
    all_data = list()
    
    for idx, folder_name in enumerate(folder_names):
        data = load(folder_name, x_axis, y_axis)
        data[labels_name] = labels[idx]
        all_data.append(data)

    all_data = pd.concat(all_data, axis=0, ignore_index=True)

    plt.close()
    sns.set_style("darkgrid")
    sns.set(font_scale=1.2)
    sns.set_style({"font.family": "serif"})

    a = sns.lineplot(
        data=all_data,
        x = x_axis,
        y = y_axis,
        hue = labels_name
        )
    
    sns.set(font_scale=1.4)
    a.set_title(title)
    plt.tight_layout()
    
    if not file_name:
        file_name = title.replace(' ', '_') + '.pdf'
    plt.savefig(file_name)
    print(f"Saved figure to {file_name}.")
    
def plot_fitness_comparison(folder_names: list, labels: list, labels_name: str = 'Labels', x_axis: str = 'Generations', y_axis: str = 'Fitness', title: str = 'Fitness comparison', file_name = None):
    '''
        Plots a figure with lines for each folder (average with shades for std) with each folder from folder_names named with the corresponding labels_name.
        In this case, the fitness is plotted
    '''
    plot_comparison(folder_names=folder_names, labels=labels, labels_name=labels_name, x_axis=x_axis, y_axis=y_axis, title=title, file_name=file_name)
    
def plot_test_fitness_comparison(folder_names: list, labels: list, labels_name: str = 'Labels', x_axis: str = 'Generations', y_axis: str = 'Test fitness', title: str = 'Test fitness comparison', file_name = None):
    '''
        Plots a figure with lines for each folder (average with shades for std) with each folder from folder_names named with the corresponding labels_name.
        In this case, the test fitness is plotted
    '''
    plot_comparison(folder_names=folder_names, labels=labels, labels_name=labels_name, x_axis=x_axis, y_axis=y_axis, title=title, file_name=file_name)
    
def plot_nodes_comparison(folder_names: list, labels: list, labels_name: str = 'Labels', x_axis: str = 'Generations', y_axis: str = 'Nodes', title: str = 'Nodes comparison', file_name = None):
    '''
        Plots a figure with lines for each folder (average with shades for std) with each folder from folder_names named with the corresponding labels_name.
        In this case, the nodes are plotted
    '''
    plot_comparison(folder_names=folder_names, labels=labels, labels_name=labels_name, x_axis=x_axis, y_axis=y_axis, title=title, file_name=file_name)

def plot_prods_comparison(folder_name: str, x_axis: str = 'Generations', extra: str = 'productions', y_axis: str = 'Fitness', title: str = 'Production comparison', file_name = None, take_out_prods: list = [ 'str', 'float', 'int' ], keep_in_prods: list | None = None):
    '''
        Plots a figure with lines for each production (average with shades for std) in the grammar (you can use take_out_prods and keep_in_prods to take out and keep in prods). Only a single folder can be given.
    '''
    
    all_data = list()
    
    data = load_w_extra(folder_name, x_axis, y_axis, extra)
    prods = data[[extra]].values[0][0].split('<class \'')[1:]
    prods = list(map(lambda x: x.split('\'>:')[0], prods))
    def obtain_value(dictionary, prod):
        only_end = dictionary.split(prod+'\'>: ')[1]
        only_beginning = only_end.split(',')[0]
        try:
            return int(only_beginning)
        except:
            return int(only_beginning.split('}')[0])
    for prod in prods:
        prod = prod.split('.')[-1]
        if prod in take_out_prods:
            continue
        if keep_in_prods:
            if prod not in keep_in_prods:
                continue
        new_data = data.copy(deep=True)
        new_data['Occurences'] = data[['productions']].squeeze().map(lambda x: obtain_value(x, prod))
        new_data['Production'] = prod
        all_data.append(new_data[[x_axis, 'Occurences', 'Production']])
    
    df = pd.concat(all_data, axis=0, ignore_index=True)
    
    plt.close()
    sns.set_style("darkgrid")
    sns.set(font_scale=1.2)
    sns.set_style({"font.family": "serif"})

    a = sns.lineplot(
        data = df,
        x = x_axis,
        y = 'Occurences',
        hue = 'Production'
        )
    
    sns.set(font_scale=1.4)
    a.set_title(title)
    plt.tight_layout()
    
    if not file_name:
        file_name = title.replace(' ', '_') + '.pdf'
    plt.savefig(file_name)
    print(f"Saved figure to {file_name}.")
    



from __future__ import annotations

import random
from typing import Any
from typing import Callable
from typing import Dict
from typing import get_args
from typing import Type

import z3

from geneticengine.core.grammar import Grammar
from geneticengine.core.random.sources import Source
from geneticengine.core.representations.tree import initialization_methods
from geneticengine.core.representations.tree.initialization_methods import is_metahandler
from geneticengine.metahandlers.base import MetaHandlerGenerator
from geneticengine.metahandlers.smt.lang import dLit
from geneticengine.metahandlers.smt.lang import dNode
from geneticengine.metahandlers.smt.lang import dVar
from geneticengine.metahandlers.smt.parser import p_expr


def simplify_type(t: type) -> type:
    if is_metahandler(t):
        return get_args(t)[0]
    return t


class SMT(MetaHandlerGenerator):
    def __init__(self, restriction_as_str="true==true"):
        self.restriction_as_str = restriction_as_str
        self.restriction = p_expr(restriction_as_str)

    def generate(
        self,
        r: Source,
        g: Grammar,
        rec,
        new_symbol,
        depth: int,
        base_type,
        context: dict[str, str],
    ):
        # fix_types(self.restriction, context)
        c = context.copy()

        ident = c["_"]
        initialization_methods.SMTResolver.register_type(ident, base_type)

        initialization_methods.SMTResolver.add_clause(
            [lambda types: self.restriction.translate(c, types)],
            {},
        )

        if base_type == int or base_type == bool or base_type == float:
            # we need the result, add receiver
            initialization_methods.SMTResolver.add_clause(
                [],
                {ident: rec},
            )
        else:
            # just synth normally
            new_symbol(base_type, rec, depth, ident, context)

    def __repr__(self):
        return f"{self.restriction}"



from __future__ import annotations

import numpy as np


def mae(y_pred, y_gt):
    """
    Calculate mean absolute error

    y_pred: the predicted output
    y_gt: ground truth
    :return: The mean absolute error.
    """

    return np.mean(np.abs(y_pred - y_gt))


def precision(y_pred, y_gt, binary=False):
    """
    Calculate precision. A higher precision corresponds to a better prediction

    y_pred: the predicted output
    y_gt: ground truth
    binary: Set to true if y_pred is already binarised
    :return: The precision.
    """
    y_pred = np.array(y_pred)
    y_gt = np.array(y_gt)
    if not binary:
        y_pred[y_pred < 0] = -1
        y_pred[y_pred >= 0] = 1

    n_positives = len(y_pred[y_pred == 1])

    fp = sum((y_pred[y_pred == 1] - y_gt[y_pred == 1]) / 2)
    precision = (n_positives - fp) / n_positives

    return precision


def recall(y_pred, y_gt, binary=False):
    """
    Calculate recall. A higher recall corresponds to a better prediction

    y_pred: the predicted output
    y_gt: ground truth
    binary: Set to true if y_pred is already binarised
    :return: The recall.
    """
    y_pred = np.array(y_pred)
    y_gt = np.array(y_gt)
    if not binary:
        y_pred[y_pred < 0] = -1
        y_pred[y_pred >= 0] = 1

    n_predicitons = len(y_pred)
    n_positives = len(y_pred[y_pred == 1])

    tp = n_positives - sum((y_pred[y_pred == 1] - y_gt[y_pred == 1]) / 2)
    fn = abs(sum((y_pred[y_pred == -1] - y_gt[y_pred == -1]) / 2))

    recall = tp / (tp + fn)
    return recall


def f1_score(y_pred, y_gt, binary=False):
    """
    Calculate f1_score. A higher f1_score corresponds to a better prediction

    y_pred: the predicted output
    y_gt: ground truth
    binary: Set to true if y_pred is already binarised
    :return: The f1_score.
    """
    y_pred = np.array(y_pred)
    y_gt = np.array(y_gt)
    if not binary:
        y_pred[y_pred < 0] = -1
        y_pred[y_pred >= 0] = 1

    n_positives = len(y_pred[y_pred == 1])

    assert y_gt.ndim == 1

    fp = sum((y_pred[y_pred == 1] - y_gt[y_pred == 1]) / 2)
    fn = abs(sum((y_pred[y_pred == -1] - y_gt[y_pred == -1]) / 2))

    tp = n_positives - fp

    denom = tp + ((fn + fp) / 2)
    if denom == 0:
        return 0

    f1_score = tp / denom
    return f1_score


def mse(y_pred, y_gt):
    """
    Calculate mean squared error. A higher mean squared error corresponds to a better prediction

    y_pred: the predicted output
    y_gt: ground truth
    binary: Set to true if y_pred is already binarised
    :return: The mean squared error.
    """

    return np.mean(np.square(y_pred - y_gt))


def rmse(y_pred, y_gt):
    """
    Calculate root mean squared error. A higher root mean squared error corresponds to a better prediction

    y_pred: the predicted output
    y_gt: ground truth
    binary: Set to true if y_pred is already binarised
    :return: The root mean squared error.
    """

    return np.sqrt(np.mean(np.square(y_pred - y_gt)))


def r2(y_pred, y_gt):
    """
    Find the theory at https://scikit-learn.org/stable/modules/generated/sklearn.linear_model.LinearRegression.html#sklearn.linear_model.LinearRegression.score
    """

    u = ((y_gt - y_pred) ** 2).sum()
    v = ((y_gt - y_gt.mean()) ** 2).sum()

    return 1 - (u / v)



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from math import isnan
from typing import Annotated
from typing import Any
from typing import Callable

import numpy as np

from geneticengine.core.tree import TreeNode
from geneticengine.grammars.sgp import Number
from geneticengine.metahandlers.ints import IntRange
from geneticengine.metahandlers.vars import VarRange


@dataclass
class SafeDiv(Number):
    """
    Safe Division object. If division fails because of a ZeroDivisionError, 1 is returned.

    Args:
        left  (Number)
        right (Number)

    Returns when evaluated:
        left / right
    """

    left: Number
    right: Number

    def evaluate(self, **kwargs):
        d1 = self.left.evaluate(**kwargs)
        d2 = self.right.evaluate(**kwargs)
        if hasattr(d1, "dtype"):
            if d1.dtype == "O":
                d1 = d1.astype(float)
        if hasattr(d2, "dtype"):
            if d2.dtype == "O":
                d2 = d2.astype(float)
        try:
            with np.errstate(divide="ignore", invalid="ignore"):
                return np.where(abs(d2) < 0.001, np.ones_like(d1), d1 / d2)
        except ZeroDivisionError:
            # In this case we are trying to divide two constants, one of which is 0
            # Return a constant.
            return 1.0

    def __str__(self) -> str:
        return f"({self.left}/{self.right})"


@dataclass
class SafeSqrt(Number):
    """
    Safe Square Root object. If the number is negative, the square root of the positive counterpart of the number is returned.

    Args:
        number (Number)

    Returns when evaluated:
        np.sqrt(number)
    """

    number: Number

    def evaluate(self, **kwargs):
        v = self.number.evaluate(**kwargs)
        return np.sqrt(np.abs(v))

    def __str__(self) -> str:
        return f"np.sqrt(np.abs({self.number}))"


@dataclass
class Sin(Number):
    """
    Standard Sinus object.

    Args:
        number (Number)

    Returns when evaluated:
        np.sin(number)
    """

    number: Number

    def evaluate(self, **kwargs):
        v = self.number.evaluate(**kwargs)
        return np.sin(v)

    def __str__(self) -> str:
        return f"np.sin({self.number})"


@dataclass
class Tanh(Number):
    """
    Standard Hyperbolic Tangent object.

    Args:
        number (Number)

    Returns when evaluated:
        np.tanh(number)
    """

    number: Number

    def evaluate(self, **kwargs):
        v = self.number.evaluate(**kwargs)
        return np.tanh(v)

    def __str__(self) -> str:
        return f"np.tanh({self.number})"


@dataclass
class Exp(Number):
    """
    Standard Exponential object.

    Args:
        number (Number)

    Returns when evaluated:
        np.exp(number)
    """

    number: Number

    def evaluate(self, **kwargs):
        v = self.number.evaluate(**kwargs)
        return np.exp(v)

    def __str__(self) -> str:
        return f"np.exp({self.number})"


@dataclass
class SafeLog(Number):
    """
    Safe Logarithmic object. If the number is negative, the logarithm of the positive counterpart of the number + 1 is returned.

    Args:
        left  (Number)
        right (Number)

    Returns when evaluated:
        np.log(number)
    """

    number: Number

    def evaluate(self, **kwargs):
        v = self.number.evaluate(**kwargs)
        return np.log(1 + np.abs(v))

    def __str__(self) -> str:
        return f"np.log(1 + np.abs({self.number}))"



from typing import List, Callable, Any


class Evaluator(object):
    def eval(self, f: Callable[[Any], float], indivs: List[Any]):
        raise NotImplementedError()


class SeqEvaluator(Evaluator):
    def eval(self, f: Callable[[Any], float], indivs: List[Any]):
        return list(map(f, indivs))



from __future__ import annotations

from typing import Callable
from typing import Optional
from typing import Type

from geneticengine.algorithms.gp.individual import Individual
from geneticengine.core.grammar import Grammar
from geneticengine.core.random.sources import Source
from geneticengine.core.representations.api import Representation


def create_mutation(
    r: Source,
    representation: Representation,
    g: Grammar,
    max_depth: int,
    specific_type: type | None = None,
    depth_aware_mut: bool = False,
) -> Callable[[Individual], Individual]:
    def mutation(individual: Individual):
        new_individual = Individual(
            genotype=representation.mutate_individual(
                r,
                g,
                individual.genotype,
                max_depth,
                g.starting_symbol,
                specific_type=specific_type,
                depth_aware_mut=depth_aware_mut,
            ),
            fitness=None,
        )
        return new_individual

    return mutation


def create_hill_climbing_mutation(
    r: Source,
    representation: Representation,
    g: Grammar,
    max_depth: int,
    fitness_function: Callable[[Individual], float],
    n_candidates: int = 5,
    specific_type: type | None = None,
    depth_aware_mut: bool = False,
) -> Callable[[Individual], Individual]:
    def hill_climbing_mutation(individual: Individual):
        def creation_new_individual():
            return Individual(
                genotype=representation.mutate_individual(
                    r,
                    g,
                    individual.genotype,
                    max_depth,
                    g.starting_symbol,
                    specific_type=specific_type,
                    depth_aware_mut=depth_aware_mut,
                ),
                fitness=None,
            )

        new_individuals = [creation_new_individual() for _ in range(n_candidates)]
        best_individual = min(
            (new_individuals + [individual]),
            key=fitness_function,
        )
        return best_individual

    return hill_climbing_mutation



from __future__ import annotations

from typing import Any
from typing import Dict
from typing import List
from typing import Protocol
from typing import runtime_checkable

from geneticengine.core.utils import get_arguments


@runtime_checkable
class TreeNode(Protocol):
    gengy_labeled: bool
    gengy_distance_to_term: int
    gengy_nodes: int
    gengy_weighted_nodes: int
    gengy_types_this_way: dict[type, list[Any]]
    gengy_init_values: tuple[Any]


class PrettyPrintable:
    def __repr__(self):
        args = ", ".join(
            [f"{a}={getattr(self, a)}" for (a, at) in get_arguments(self.__class__)],
        )
        return f"{self.__class__.__name__}({args})"



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import TypeVar

from geneticengine.core.grammar import Grammar
from geneticengine.core.random.sources import Source
from geneticengine.core.representations.api import Representation
from geneticengine.core.representations.tree.initialization_methods import Initialization_Method, PI_Grow, Random_Production, mk_save_init
from geneticengine.core.representations.tree.utils import GengyList
from geneticengine.core.representations.tree.utils import relabel_nodes
from geneticengine.core.representations.tree.utils import relabel_nodes_of_trees
from geneticengine.core.tree import TreeNode
from geneticengine.core.utils import get_arguments
from geneticengine.core.utils import get_generic_parameter
from geneticengine.core.utils import has_annotated_crossover
from geneticengine.core.utils import has_annotated_mutation
from geneticengine.core.utils import is_abstract
from geneticengine.exceptions import GeneticEngineError

T = TypeVar("T")




def random_node(
    r: Source,
    g: Grammar,
    max_depth: int,
    starting_symbol: type[Any] | None = None,
    method=PI_Grow().tree_init_method,
):
    if starting_symbol is None:
        starting_symbol = g.starting_symbol
    return method(r, g, max_depth, starting_symbol)




def random_individual(r: Source, g: Grammar, max_depth: int = 5, method = PI_Grow().tree_init_method) -> TreeNode:
    try:
        assert max_depth >= g.get_min_tree_depth()
    except:
        if g.get_min_tree_depth() == 1000000:
            raise GeneticEngineError(
                f"Grammar's minimal tree depth is {g.get_min_tree_depth()}, which is the default tree depth. It's highly like that there are nodes of your grammar than cannot reach any terminal.",
            )
        raise GeneticEngineError(
            f"Cannot use complete grammar for individual creation. Max depth ({max_depth}) is smaller than grammar's minimal tree depth ({g.get_min_tree_depth()}).",
        )
    ind = random_node(r, g, max_depth, g.starting_symbol, method)
    assert isinstance(ind, TreeNode)
    return ind


def mutate_inner(
    r: Source,
    g: Grammar,
    i: TreeNode,
    max_depth: int,
    ty: type,
    force_mutate: bool,
    depth_aware_mut: bool,
) -> TreeNode:
    counter = i.gengy_weighted_nodes if depth_aware_mut else i.gengy_nodes
    if counter > 0:
        c = r.randint(0, counter - 1)
        if (
            c == 0
            or (c <= i.gengy_distance_to_term and depth_aware_mut)
            or force_mutate
        ):
            # If Metahandler mutation exists, the mutation process is different
            args_with_specific_mutation = [
                has_annotated_mutation(arg[1]) for arg in get_arguments(i)
            ]
            if any(args_with_specific_mutation):
                mutation_possibilities = len(args_with_specific_mutation)
                mutation_choice = r.randint(
                    0,
                    mutation_possibilities - 1,
                )
                (index, arg_to_be_mutated) = [
                    (kdx, arg[1])
                    for kdx, arg in enumerate(get_arguments(i))
                    if args_with_specific_mutation[kdx]
                ][mutation_choice]
                args = list(i.gengy_init_values)
                args[index] = arg_to_be_mutated.__metadata__[0].mutate(  # type: ignore
                    r,
                    g,
                    random_node,
                    max_depth - 1,
                    get_generic_parameter(arg_to_be_mutated),
                    method=Random_Production().tree_init_method,
                    current_node=args[index],
                )
                mk = mk_save_init(type(i), lambda x: x)(*args)
                return mk

            for _ in range(5):
                replacement = random_node(r, g, max_depth, ty, method=Random_Production().tree_init_method)
                if replacement != i:
                    break
            return replacement
        else:
            if is_abstract(ty) and g.expansion_depthing:
                max_depth -= g.abstract_dist_to_t[ty][type(i)]
            max_depth -= 1
            args = list(i.gengy_init_values)
            c -= i.gengy_distance_to_term if depth_aware_mut else 1
            for idx, (_, field_type) in enumerate(get_arguments(i)):
                child = args[idx]
                if hasattr(child, "gengy_nodes"):
                    count = (
                        child.gengy_weighted_nodes
                        if depth_aware_mut
                        else child.gengy_nodes
                    )
                    if c <= count:
                        mi = mutate_inner(
                            r,
                            g,
                            child,
                            max_depth,
                            field_type,
                            force_mutate,
                            depth_aware_mut,
                        )
                        args[idx] = mi
                        break
                    else:
                        c -= count
            mk = mk_save_init(i, lambda x: x)(*args)
            return mk
    else:
        for _ in range(5):
            rn = random_node(r, g, max_depth, ty, method=Random_Production().tree_init_method)
            if rn != i:
                break
        return rn


def mutate_specific_type_inner(
    r: Source,
    g: Grammar,
    i: TreeNode,
    max_depth: int,
    ty: type,
    specific_type: type,
    n: int,
    depth_aware_mut: bool,
) -> TreeNode:
    if n == 1 and type(i) == specific_type:
        return mutate_inner(
            r,
            g,
            i,
            max_depth,
            ty,
            force_mutate=True,
            depth_aware_mut=depth_aware_mut,
        )
    else:
        args = list(i.gengy_init_values)
        for idx, (_, field_type) in enumerate(get_arguments(i)):
            child = args[idx]
            n_options = len(
                list(find_in_tree_exact(g, specific_type, child, max_depth)),
            )
            if n_options <= n:
                args[idx] = mutate_specific_type_inner(
                    r,
                    g,
                    child,
                    max_depth,
                    ty,
                    specific_type,
                    n,
                    depth_aware_mut,
                )
            else:
                n -= n_options
        return mk_save_init(i, lambda x: x)(*args)


def mutate_specific_type(
    r: Source,
    g: Grammar,
    i: TreeNode,
    max_depth: int,
    target_type: type,
    specific_type: type,
    depth_aware_mut: bool,
) -> TreeNode:
    ch = r.randint(0, 2)
    n_options = len(list(find_in_tree_exact(g, specific_type, i, max_depth)))
    if ch == 0 or n_options == 0:
        new_tree = mutate_inner(
            r,
            g,
            i,
            max_depth,
            target_type,
            force_mutate=False,
            depth_aware_mut=depth_aware_mut,
        )
        relabeled_new_tree = relabel_nodes_of_trees(new_tree, g)
        return relabeled_new_tree
    else:
        n = r.randint(1, n_options)
        new_tree = mutate_specific_type_inner(
            r,
            g,
            i,
            max_depth,
            target_type,
            specific_type,
            n,
            depth_aware_mut,
        )
        relabeled_new_tree = relabel_nodes_of_trees(new_tree, g)
        return relabeled_new_tree


def mutate(
    r: Source,
    g: Grammar,
    i: TreeNode,
    max_depth: int,
    target_type: type,
    specific_type: type | None = None,
    depth_aware_mut: bool = False,
) -> Any:
    if specific_type:
        return mutate_specific_type(
            r,
            g,
            i,
            max_depth,
            target_type,
            specific_type,
            depth_aware_mut,
        )
    new_tree = mutate_inner(r, g, i, max_depth, target_type, False, depth_aware_mut)
    relabeled_new_tree = relabel_nodes_of_trees(new_tree, g)
    return relabeled_new_tree


def find_in_tree(g: Grammar, ty: type, o: TreeNode, max_depth: int):
    is_abs = is_abstract(ty)
    if hasattr(o, "gengy_types_this_way"):

        for t in o.gengy_types_this_way:

            def is_valid(node):
                _, depth, _, _ = relabel_nodes(node, g)

                if is_abs and g.expansion_depthing:
                    depth += g.abstract_dist_to_t[ty][t]

                return depth <= max_depth

            if ty in t.__bases__:
                vals = o.gengy_types_this_way[t]
                if vals:
                    yield from filter(is_valid, vals)


def find_in_tree_exact(g: Grammar, ty: type, o: TreeNode, max_depth: int):
    if hasattr(o, "gengy_types_this_way"):

        vals = o.gengy_types_this_way[ty]
        if vals:

            def is_valid(node):
                _, depth, _, _ = relabel_nodes(node, g)
                return depth <= max_depth

            yield from filter(is_valid, vals)


def crossover_inner(
    r: Source,
    g: Grammar,
    i: TreeNode,
    o: TreeNode,
    max_depth: int,
    ty: type,
    force_crossover: bool,
    depth_aware_co: bool,
) -> Any:
    counter = i.gengy_weighted_nodes if depth_aware_co else i.gengy_nodes
    if counter > 0:
        c = r.randint(0, counter - 1)
        if (
            c == 0
            or (c <= i.gengy_distance_to_term and depth_aware_co)
            or force_crossover
        ):
            replacement = None
            args_with_specific_crossover = [
                has_annotated_crossover(arg[1]) for arg in get_arguments(i)
            ]
            if any(args_with_specific_crossover):
                crossover_possibilities = len(args_with_specific_crossover)
                crossover_choice = r.randint(
                    0,
                    crossover_possibilities - 1,
                )
                options = list(find_in_tree_exact(g, type(i), o, max_depth))
                if not options:
                    pass  # Replace whole node
                else:
                    (index, arg_to_be_crossovered) = [
                        (kdx, arg)
                        for kdx, arg in enumerate(get_arguments(i))
                        if args_with_specific_crossover[kdx]
                    ][crossover_choice]
                    args = list(i.gengy_init_values)
                    args[index] = (
                        arg_to_be_crossovered[1]
                        .__metadata__[0]  # type: ignore
                        .crossover(
                            r,
                            g,
                            options,
                            arg_to_be_crossovered[0],
                            ty,
                            current_node=args[index],
                        )
                    )
                    return mk_save_init(type(i), lambda x: x)(*args)

            options = list(find_in_tree(g, ty, o, max_depth))
            if options:
                replacement = r.choice(options)
            if replacement is None:
                for _ in range(5):
                    replacement = random_node(r, g, max_depth, ty, method=Random_Production().tree_init_method)
                    if replacement != i:
                        break

            return replacement
        else:
            if is_abstract(ty) and g.expansion_depthing:
                max_depth -= g.abstract_dist_to_t[ty][type(i)]
            max_depth -= 1
            args = list(i.gengy_init_values)
            c -= i.gengy_distance_to_term if depth_aware_co else 1
            for idx, (field, field_type) in enumerate(get_arguments(i)):
                child = args[idx]
                if hasattr(child, "gengy_nodes"):
                    count = (
                        child.gengy_weighted_nodes
                        if depth_aware_co
                        else child.gengy_nodes
                    )
                    if c <= count:
                        args[idx] = crossover_inner(
                            r,
                            g,
                            child,
                            o,
                            max_depth,
                            field_type,
                            force_crossover=False,
                            depth_aware_co=depth_aware_co,
                        )
                        break
                    else:
                        c -= count
            return mk_save_init(i, lambda x: x)(*args)
    else:
        return i


def crossover_specific_type_inner(
    r: Source,
    g: Grammar,
    i: TreeNode,
    o: TreeNode,
    max_depth: int,
    ty: type,
    specific_type: type,
    n: int,
    depth_aware_co: bool,
) -> TreeNode:
    if n == 1 and type(i) == specific_type:
        return crossover_inner(
            r,
            g,
            i,
            o,
            max_depth,
            ty,
            force_crossover=True,
            depth_aware_co=depth_aware_co,
        )
    else:
        args = list(i.gengy_init_values)
        for idx, (_, field_type) in enumerate(get_arguments(i)):
            child = args[idx]
            n_options = len(
                list(find_in_tree_exact(g, specific_type, child, max_depth)),
            )
            if n_options <= n:
                args[idx] = crossover_specific_type_inner(
                    r,
                    g,
                    child,
                    o,
                    max_depth,
                    ty,
                    specific_type,
                    n,
                    depth_aware_co=depth_aware_co,
                )
            else:
                n -= n_options
        return mk_save_init(i, lambda x: x)(*args)


def crossover_specific_type(
    r: Source,
    g: Grammar,
    i: TreeNode,
    o: TreeNode,
    max_depth: int,
    target_type: type,
    specific_type: type,
    depth_aware_co: bool,
) -> TreeNode:
    ch = r.randint(0, 1)
    n_options_i = len(list(find_in_tree_exact(g, specific_type, i, max_depth)))
    n_options_o = len(list(find_in_tree_exact(g, specific_type, o, max_depth)))
    if ch == 0 or n_options_i == 0 or n_options_o == 0:
        new_tree = crossover_inner(
            r,
            g,
            i,
            o,
            max_depth,
            target_type,
            force_crossover=False,
            depth_aware_co=depth_aware_co,
        )
        relabeled_new_tree = relabel_nodes_of_trees(new_tree, g)
        return relabeled_new_tree
    else:
        n = r.randint(1, n_options_i)
        new_tree = crossover_specific_type_inner(
            r,
            g,
            i,
            o,
            max_depth,
            target_type,
            specific_type,
            n,
            depth_aware_co=depth_aware_co,
        )
        relabeled_new_tree = relabel_nodes_of_trees(new_tree, g)
        return relabeled_new_tree


def crossover(
    r: Source,
    g: Grammar,
    p1: TreeNode,
    p2: TreeNode,
    max_depth: int,
    specific_type: type | None = None,
    depth_aware_co: bool = False,
) -> tuple[TreeNode, TreeNode]:
    """
    Given the two input trees [p1] and [p2], the grammar and the random source, this function returns two trees that are created by crossing over [p1] and [p2]. The first tree returned has [p1] as the base, and the second tree has [p2] as a base.
    """
    if specific_type:
        new_tree1 = crossover_specific_type(
            r,
            g,
            p1,
            p2,
            max_depth,
            g.starting_symbol,
            specific_type,
            depth_aware_co=depth_aware_co,
        )
    else:
        new_tree1 = crossover_inner(
            r,
            g,
            p1,
            p2,
            max_depth,
            g.starting_symbol,
            force_crossover=False,
            depth_aware_co=depth_aware_co,
        )
    relabeled_new_tree1 = relabel_nodes_of_trees(new_tree1, g)

    if specific_type:
        new_tree2 = crossover_specific_type(
            r,
            g,
            p2,
            p1,
            max_depth,
            g.starting_symbol,
            specific_type,
            depth_aware_co=depth_aware_co,
        )
    else:
        new_tree2 = crossover_inner(
            r,
            g,
            p2,
            p1,
            max_depth,
            g.starting_symbol,
            force_crossover=False,
            depth_aware_co=depth_aware_co,
        )
    relabeled_new_tree2 = relabel_nodes_of_trees(new_tree2, g)
    return relabeled_new_tree1, relabeled_new_tree2


class TreeBasedRepresentation(Representation[TreeNode]):
    """This class represents the tree representation of an individual.
    In this approach, the genotype and the phenotype are exactly the same."""

    def __init__(self, depth = None, method: Initialization_Method = PI_Grow()) -> None:
        self.depth = depth
        self.method = method
        
    def create_individual(self, r: Source, g: Grammar, depth: int) -> TreeNode:
        return random_individual(r, g, depth, self.method.tree_init_method)

    def mutate_individual(
        self,
        r: Source,
        g: Grammar,
        ind: TreeNode,
        depth: int,
        ty: type,
        specific_type: type | None = None,
        depth_aware_mut: bool = False,
    ) -> TreeNode:
        new_ind = mutate(r, g, ind, depth, ty, specific_type, depth_aware_mut)
        return new_ind

    def crossover_individuals(
        self,
        r: Source,
        g: Grammar,
        i1: TreeNode,
        i2: TreeNode,
        max_depth: int,
        specific_type: type | None = None,
        depth_aware_co: bool = False,
    ) -> tuple[TreeNode, TreeNode]:
        return crossover(r, g, i1, i2, max_depth, specific_type, depth_aware_co)

    def genotype_to_phenotype(self, g: Grammar, genotype: TreeNode) -> TreeNode:
        return genotype


treebased_representation = TreeBasedRepresentation()



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Dict
from typing import Type

from geneticengine.core.grammar import Grammar
from geneticengine.core.random.sources import Source
from geneticengine.metahandlers.base import MetaHandlerGenerator


class WeightedStringHandler(MetaHandlerGenerator):
    """
    This metahandler restricts the creation of string nodes
    so that the output complies with a given alphabet and a
    matrix of probabilities for each position.

    Each row on the matrix should reflect the probability of
    each character in that position. Thus, the number of cols
    in the input matrix should be the same as the number of
    characters in the alphabet.

    This refinement will return a string with a
    size == nrows in the matrix
    """

    def __init__(self, matrix, alphabet):
        self.probability_matrix = matrix
        self.alphabet = alphabet

        assert (
            len(self.alphabet) == self.probability_matrix.shape[1]
        ), "Cols in probability matrix must have the same size as the alphabet provided"

    def generate(
        self,
        r: Source,
        g: Grammar,
        rec,
        newsymbol,
        depth: int,
        base_type,
        context: dict[str, str],
    ):
        out = ""
        for row in self.probability_matrix:
            out += r.choice_weighted(self.alphabet, row, str(base_type))
        rec(out)

    def __repr__(self):
        return f"str[aphabet={self.alphabet}, size={self.probability_matrix.shape[0]}"

    def __class_getitem__(self, args):
        return WeightedStringHandler(*args)



from __future__ import annotations

from dataclasses import dataclass

from geneticengine.core.decorators import abstract
from geneticengine.grammars.sgp import Number


@abstract
class ExpLiteral(Number):
    pass


@dataclass
class One(ExpLiteral):
    def evaluate(self, **kwargs):
        return 1

    def __str__(self) -> str:
        return str(1)


@dataclass
class PointOne(ExpLiteral):
    def evaluate(self, **kwargs):
        return 0.1

    def __str__(self) -> str:
        return str(0.1)


@dataclass
class PointtOne(ExpLiteral):
    def evaluate(self, **kwargs):
        return 0.01

    def __str__(self) -> str:
        return str(0.01)


@dataclass
class PointttOne(ExpLiteral):
    def evaluate(self, **kwargs):
        return 0.001

    def __str__(self) -> str:
        return str(0.001)


@dataclass
class MinusPointttOne(ExpLiteral):
    def evaluate(self, **kwargs):
        return -0.001

    def __str__(self) -> str:
        return str(-0.001)


@dataclass
class MinusPointtOne(ExpLiteral):
    def evaluate(self, **kwargs):
        return -0.01

    def __str__(self) -> str:
        return str(-0.01)


@dataclass
class MinusPointOne(ExpLiteral):
    def evaluate(self, **kwargs):
        return -0.1

    def __str__(self) -> str:
        return str(-0.1)


@dataclass
class MinusOne(ExpLiteral):
    def evaluate(self, **kwargs):
        return -1

    def __str__(self) -> str:
        return str(-1)


exp_literals = [
    MinusOne,
    MinusPointOne,
    MinusPointtOne,
    MinusPointttOne,
    One,
    PointOne,
    PointtOne,
    PointttOne,
]



from __future__ import annotations

from dataclasses import dataclass
from textwrap import indent
from typing import Annotated
from typing import Any
from typing import Callable
from typing import List

from geneticengine.core.representations.tree.utils import GengyList
from geneticengine.grammars.coding.classes import Condition
from geneticengine.grammars.coding.classes import Statement
from geneticengine.metahandlers.ints import IntRange
from geneticengine.metahandlers.lists import ListSizeBetween


@dataclass
class Code(Statement):
    stmts: Annotated[list[Statement], ListSizeBetween(1, 4)]

    def evaluate(self, **kwargs) -> float:
        x = kwargs.get("x", 1.0)
        for stmt in self.stmts:
            x = stmt.evaluate(x=x)
        return x

    def evaluate_lines(self, **kwargs) -> Callable[[Any], float]:
        def ev(line):
            for stmt in self.stmts:
                x = stmt.evaluate_lines(**kwargs)(line)
            return x

        return lambda line: ev(line)

    def __str__(self):
        return "\n".join([str(stmt) for stmt in self.stmts])


@dataclass
class ForLoop(Statement):
    iterationRange: Annotated[int, IntRange(1, 6)]
    loopedCode: Statement

    def evaluate(self, **kwargs) -> float:
        x = kwargs.get("x", 1.0)
        assert type(self.iterationRange) == int
        for _ in range(self.iterationRange):
            x = self.loopedCode.evaluate(x=x)
        return x

    def evaluate_lines(self, **kwargs) -> Callable[[Any], float]:
        def ev(line):
            for _ in range(self.iterationRange):
                x = self.loopedCode.evaluate_lines(**kwargs)(line)
            return x

        return lambda line: ev(line)

    def __str__(self):
        return "for i in range({}):\n{}".format(
            self.iterationRange,
            indent(str(self.loopedCode), "\t"),
        )


@dataclass
class While(Statement):
    # Shouldn't be used! We don't have a stopping criteria, usage will always result in infinite loops
    cond: Condition
    loopedCode: Statement

    def evaluate(self, **kwargs) -> float:
        x = kwargs.get("x", 1.0)
        for _ in range(self.cond.evaluate(x=x)):
            x = self.loopedCode.evaluate(x=x)
        return x

    def evaluate_lines(self, **kwargs) -> Callable[[Any], float]:
        def ev(line):
            while self.cond.evaluate_lines(**kwargs)(line):
                x = self.loopedCode.evaluate_lines(**kwargs)(line)
            return x

        return lambda line: ev(line)

    def __str__(self):
        return "while {}:\n{}".format(self.cond, indent(str(self.loopedCode), "\t"))


@dataclass
class IfThen(Statement):
    cond: Condition
    then: Statement

    def evaluate(self, **kwargs) -> float:
        x = kwargs.get("x", 1.0)
        if self.cond.evaluate(x=x):
            x = self.then.evaluate(x=x)
        return x

    def evaluate_lines(self, **kwargs) -> Callable[[Any], float]:
        def ev(line):
            if self.cond.evaluate_lines(**kwargs)(line):
                x = self.then.evaluate_lines(**kwargs)(line)
            else:
                x = 0  # Is this correctly done?
            return x

        return lambda line: ev(line)

    def __str__(self):
        return "if {}:\n{}".format(self.cond, indent(str(self.then), "\t"))


@dataclass
class IfThenElse(Statement):
    cond: Condition
    then: Statement
    elze: Statement

    def evaluate(self, **kwargs) -> float:
        x = kwargs.get("x", 1.0)
        if self.cond.evaluate(x=x):
            x = self.then.evaluate(x=x)
        else:
            x = self.elze.evaluate(x=x)
        return x

    def evaluate_lines(self, **kwargs) -> Callable[[Any], float]:
        def ev(line):
            if self.cond.evaluate_lines(**kwargs)(line):
                z = self.then.evaluate_lines(**kwargs)(line)
            else:
                z = self.elze.evaluate_lines(**kwargs)(line)
            return z

        return lambda line: ev(line)

    def __str__(self):
        return "if {}:\n{}\nelse:\n{}".format(
            self.cond,
            indent(str(self.then), "\t"),
            indent(
                str(self.elze),
                "\t",
            ),
        )



from __future__ import annotations

import inspect
from abc import ABC
from dataclasses import dataclass
from functools import wraps
from typing import Any
from typing import Callable
from typing import get_type_hints
from typing import List
from typing import Protocol
from typing import Set
from typing import Tuple
from typing import Type
from typing import TYPE_CHECKING

from geneticengine.algorithms.gp.individual import Individual
from geneticengine.core.decorators import get_gengy

if TYPE_CHECKING:
    from geneticengine.core.representations.tree.utils import GengyList
    from geneticengine.core.tree import TreeNode


def has_annotated_mutation(ty: type[Any]):
    """Returns whether type has an annotated mutation within metadata."""
    if hasattr(ty, "__metadata__"):
        if hasattr(ty.__metadata__[0], "mutate"):
            return True
    return False


def has_annotated_crossover(ty: type[Any]):
    """Returns whether type has an annotated crossover within metadata."""
    if hasattr(ty, "__metadata__"):
        if hasattr(ty.__metadata__[0], "crossover"):
            return True
    return False


def is_annotated(ty: type[Any]):
    """Returns whether type is annotated with metadata."""
    return hasattr(ty, "__metadata__")


def is_generic_list(ty: type[Any]):
    """Returns whether a type is List[T] for any T"""
    return hasattr(ty, "__origin__") and ty.__origin__ is list


def is_generic(ty: type[Any]):
    """Returns whether a type is x[T] for any T"""
    return hasattr(ty, "__origin__")


def get_generic_parameters(ty: type[Any]) -> list[type]:
    """Annotated[T, <annotations>] or List[T], this function returns Dict[T,]"""
    return ty.__args__


def get_generic_parameter(ty: type[Any]) -> type:
    """When given Annotated[T, <annotations>] or List[T], this function returns T"""
    return get_generic_parameters(ty)[0]


def strip_annotations(ty: type[Any]) -> type:
    """When given Annotated[T, <annotations>] or List[T], this function recurses with T
    Otherwise, it returns the parameter unchanged.
    """
    if is_generic_list(ty) or is_annotated(ty):
        return strip_annotations(get_generic_parameter(ty))
    else:
        return ty


def has_arguments(n: Any) -> bool:
    """Returns whether a node has arguments or not"""
    return (
        hasattr(n, "__init__")
        and hasattr(n.__init__, "__annotations__")
        and len(n.__init__.__annotations__) > 0
    )


def get_arguments(n) -> list[tuple[str, type]]:
    """
    :param n: production
    :return: list((argname, argtype))
    """
    if hasattr(n, "__init__"):
        init = n.__init__
        import sys

        args: dict[str, type] = get_type_hints(
            init,
            globalns=sys.modules[n.__module__].__dict__,
            include_extras=True,
        )
        return [(a, args[a]) for a in filter(lambda x: x != "return", args)]
    elif isinstance(n, GengyList):
        return [(f"{i}", n.typ) for i in range(len(n))]
    return []


def is_abstract(t: type) -> bool:
    """Returns whether a class is a Protocol or AbstractBaseClass"""
    return t.mro()[1] in [ABC, Protocol] or get_gengy(t).get("abstract", False)


def is_terminal(t: type, non_terminals: set[type]) -> bool:
    """Returns whether a node is a terminal or not, based on the list of non terminals in the grammar"""
    if is_annotated(t):
        return all(
            [is_terminal(inner, non_terminals) for inner in get_generic_parameters(t)],
        )
    return t not in non_terminals


def all_init_arguments_typed(t: type) -> bool:
    if hasattr(t, "__init__"):
        m = getattr(t, "__init__")
        d = inspect.getfullargspec(m)
        return all(x in d.annotations for x in d.args[1:])  # starts with self
    return False


# debug_fin


def build_finalizers(
    final_callback,
    n_args,
    per_callback: list[Callable[[Any], None]] | None = None,
) -> list[Any]:
    """
    Builds a set of functions that accumulate the arguments provided
    :param final_callback:
    :param n_args:
    :return:
    """
    uninit = object()
    rets = [uninit] * n_args
    to_arrive = [n_args]

    finalizers = []
    # id = debug_fin[0]
    # print("%i has %i fin " % (id, n_args))
    # debug_fin[0] += 1

    for i in range(n_args):

        def fin(x, i=i):
            if rets[i] is uninit:
                rets[i] = x
                to_arrive[0] -= 1

                if per_callback is not None:
                    per_callback[i](x)

                if to_arrive[0] == 0:
                    # we recieved all params, finish construction
                    final_callback(*rets)
                # else:
                #     print("%i prog %i" % (id, to_arrive[0]))
            else:
                raise Exception(
                    "Received duplicate param on finalizer! i=%d" % i,
                )

        finalizers.append(fin)

    if n_args == 0:
        final_callback()

    return finalizers


def average_fitness(individual: Individual) -> float:
    """Returns the average fitness list value of the given individual"""
    assert isinstance(individual.fitness, list)
    average_fitness = sum(individual.fitness) / len(individual.fitness)
    return average_fitness



from __future__ import annotations

import csv
from typing import Any, Callable
from typing import Optional

from geneticengine.algorithms.callbacks.callback import Callback
from geneticengine.algorithms.gp.individual import Individual


class CSVCallback(Callback):
    """
    Callback that outputs to a given CSV file.
    
    Args:
        filename (str | None): The file name where the results are saved to (default: evolution_results.csv).
        filter_population (Callable[[list[Individual]], list[Individual]]): The method of filtering the population before saving (default: identity function, aka: lambda x: x).
        test_data (Callable[[Any], float] | None): A fitness_function-style function to save test results (genotype -> test_fitness).
        only_record_best_ind (bool): Set to False if you want to save data on all individuals. This is memory expensive (default: True).
        save_genotype_as_string (bool): Set to True if you want to save the genotype if the individuals. This is memory expensive, especially in combination with the only_record_best_ind set to False (default: False).
        save_productions (bool): Set to True to save the occurences of productions of individuals (default: False).

    """

    def __init__(
        self,
        filename: str | None = None,
        filter_population: Callable[[list[Individual]], list[Individual]] = lambda x: x,
        test_data: Callable[[Any], float] | None = None,
        only_record_best_ind: bool = True,
        save_genotype_as_string: bool = False,
        save_productions: bool = False,
    ):
        if filename is None:
            filename = "evolution_results.csv"
        self.filename = filename
        self.filter_population = filter_population
        self.time = 0.0
        self.test_data = test_data
        self.only_record_best_ind = only_record_best_ind
        self.save_genotype_as_string = save_genotype_as_string
        self.save_productions = save_productions
        self.write_header()

    def end_evolution(self):
        self.outfile.close()

    def write_header(self):
        self.outfile = open(f"{self.filename}", "w", newline="")
        self.writer = csv.writer(self.outfile)
        row = [
            "Fitness",
            "Depth",
            "Nodes",
            "Generations",
            "time_since_the_start_of_the_evolution",
            "seed",
        ]
        if self.test_data:
            row.append("Test fitness")
        if self.save_genotype_as_string:
            row.append("genotype_as_str")
        if self.save_productions:
            row.append("productions")
        self.writer.writerow(row)

    def process_iteration(self, generation: int, population, time: float, gp):
        pop = self.filter_population(population)
        if self.only_record_best_ind:
            pop = [gp.get_best_individual(gp.problem, population)]
        self.time = time
        for ind in pop:
            phenotype = gp.representation.genotype_to_phenotype(
                        gp.grammar,
                        ind.genotype,
                    )
            if hasattr(phenotype, "gengy_distance_to_term"):
                depth = phenotype.gengy_distance_to_term
            else:
                depth = -1
            if hasattr(phenotype, "gengy_nodes"):
                nodes = phenotype.gengy_nodes
            else:
                nodes = -1
            row = [
                ind.fitness,
                depth,
                nodes,
                generation,
                self.time,
                gp.seed,
            ]
            if self.test_data:
                row.append(self.test_data(phenotype))
            if self.save_genotype_as_string:
                row.append(ind.genotype)
            if self.save_productions:
                row.append(ind.count_prods(gp.representation.genotype_to_phenotype, gp.grammar))
            self.writer.writerow([str(x) for x in row])



from __future__ import annotations

from dataclasses import dataclass
from typing import Any
from typing import List
from typing import Protocol
from typing import Tuple
from typing import Type

from geneticengine.core.grammar import Grammar
from geneticengine.core.random.sources import RandomSource
from geneticengine.core.random.sources import Source
from geneticengine.core.representations.api import Representation
from geneticengine.core.representations.tree.initialization_methods import Initialization_Method, PI_Grow
from geneticengine.core.representations.tree.treebased import random_node
from geneticengine.core.tree import TreeNode


@dataclass
class Genotype:
    dna: list[int]


def random_individual(
    r: Source,
    g: Grammar,
    depth: int = 5,
    starting_symbol: Any = None,
    gene_size: int = 256,
) -> Genotype:
    return Genotype([r.randint(0, 10000) for _ in range(gene_size)])


def mutate(r: Source, g: Grammar, ind: Genotype, max_depth: int) -> Genotype:
    rindex = r.randint(0, len(ind.dna) - 1)
    clone = [i for i in ind.dna]
    clone[rindex] = r.randint(0, 10000)
    return Genotype(clone)


def crossover(
    r: Source,
    g: Grammar,
    p1: Genotype,
    p2: Genotype,
    max_depth: int,
) -> tuple[Genotype, Genotype]:
    rindex = r.randint(0, len(p1.dna) - 1)
    c1 = p1.dna[:rindex] + p2.dna[rindex:]
    c2 = p2.dna[:rindex] + p1.dna[rindex:]
    return (Genotype(c1), Genotype(c2))


class ListWrapper(Source):
    dna: list[int]
    index: int = 0
    
    def __init__(self, dna):
        self.dna = dna

    def randint(self, min: int, max: int, prod: str = "") -> int:
        self.index = (self.index + 1) % len(self.dna)
        v = self.dna[self.index]
        return v % (max - min + 1) + min

    def random_float(self, min: float, max: float, prod: str = "") -> float:
        k = self.randint(1, 100000000, prod)
        return 1 * (max - min) / k + min


def create_tree(g: Grammar, ind: Genotype, depth: int, method) -> TreeNode:
    rand: Source = ListWrapper(ind.dna)
    return random_node(rand, g, depth, g.starting_symbol, method=method)


class GrammaticalEvolutionRepresentation(Representation[Genotype]):
    """This representation uses a list of integers to guide the generation of trees in the phenotype."""

    def __init__(self, depth = None, gene_size=256, method: Initialization_Method = PI_Grow()) -> None:
        self.depth = depth
        self.gene_size = gene_size
        self.method = method

    def create_individual(self, r: Source, g: Grammar, depth: int) -> Genotype:
        self.depth = depth
        return random_individual(r, g, depth, gene_size=self.gene_size)

    def mutate_individual(
        self,
        r: Source,
        g: Grammar,
        ind: Genotype,
        depth: int,
        ty: type,
        specific_type: type | None = None,
        depth_aware_mut: bool = False,
    ) -> Genotype:
        return mutate(r, g, ind, depth)

    def crossover_individuals(
        self,
        r: Source,
        g: Grammar,
        i1: Genotype,
        i2: Genotype,
        depth: int,
        specific_type: type | None = None,
        depth_aware_co: bool = False,
    ) -> tuple[Genotype, Genotype]:
        return crossover(r, g, i1, i2, depth)

    def genotype_to_phenotype(self, g: Grammar, genotype: Genotype) -> TreeNode:
        return create_tree(g, genotype, self.depth, self.method.tree_init_method)


ge_representation = GrammaticalEvolutionRepresentation()



from __future__ import annotations

from dataclasses import dataclass
from typing import Any
from typing import Callable

from geneticengine.grammars.coding.classes import Condition


@dataclass
class And(Condition):
    """attributes left and right"""

    left: Condition
    right: Condition

    def evaluate(self, **kwargs) -> bool:
        return self.left.evaluate(**kwargs) and self.right.evaluate(**kwargs)

    def evaluate_lines(self, **kwargs) -> Callable[[Any], bool]:
        return lambda line: self.left.evaluate_lines(**kwargs)(
            line,
        ) and self.right.evaluate_lines(**kwargs)(line)

    def __str__(self):
        return f"({self.left} and {self.right})"


@dataclass
class Or(Condition):
    """attributes left and right"""

    left: Condition
    right: Condition

    def evaluate(self, **kwargs) -> bool:
        return self.left.evaluate(**kwargs) or self.right.evaluate(**kwargs)

    def evaluate_lines(self, **kwargs) -> Callable[[Any], bool]:
        return lambda line: self.left.evaluate_lines(**kwargs)(
            line,
        ) or self.right.evaluate_lines(**kwargs)(line)

    def __str__(self):
        return f"({self.left} or {self.right})"


@dataclass
class Not(Condition):
    """attribute cond"""

    cond: Condition

    def evaluate(self, **kwargs) -> bool:
        return not self.cond.evaluate(**kwargs)

    def evaluate_lines(self, **kwargs) -> Callable[[Any], bool]:
        return lambda line: not self.cond.evaluate_lines(**kwargs)(line)

    def __str__(self):
        return f"(not {self.cond})"



from __future__ import annotations

import random
from abc import ABC, abstractmethod
from itertools import accumulate
from typing import Any
from typing import List
from typing import TypeVar

from geneticengine.core.representations.tree.utils import GengyList
from geneticengine.core.utils import build_finalizers
from geneticengine.core.utils import get_generic_parameter

T = TypeVar("T")


class Source(ABC):
    @abstractmethod
    def __init__(self, seed: int = 0):
        ...

    @abstractmethod
    def randint(self, min: int, max: int, prod: str = "") -> int:
        ...

    @abstractmethod
    def random_float(self, min: float, max: float, prod: str = "") -> float:
        ...

    def choice(self, choices: list[T], prod: str = "") -> T:
        assert choices
        i = self.randint(0, len(choices) - 1, prod)
        return choices[i]

    def choice_weighted(
        self,
        choices: list[T],
        weights: list[float],
        prod: str = "",
    ) -> T:
        acc_weights = list(accumulate(weights))
        total = acc_weights[-1] + 0.0
        rand_value: float = self.random_float(0, total, prod)

        for (choice, acc) in zip(choices, acc_weights):
            if rand_value < acc:
                return choice
        return choices[0]

    def shuffle(self, l: list[T], prod: str = ""):
        for i in reversed(range(1, len(l))):
            j = self.randint(0, i, prod)
            l[i], l[j] = l[j], l[i]
        return l

    def pop_random(self, l: list[T], prod: str = "") -> T:
        item = l.pop()
        total_len = len(l)

        i = self.randint(0, total_len, prod)
        if i == total_len:
            return item

        l[i], item = item, l[i]

        return item

    def random_bool(self, prod: str = "") -> bool:
        return self.choice([True, False], prod)

    def random_list(
        self,
        receiver,
        new_symbol,
        depth: int,
        ty: type[list[T]],
        ctx: dict[str, str],
        prod: str = "",
    ):
        inner_type = get_generic_parameter(ty)
        size = 1
        if depth > 0:
            size = self.randint(1, depth, prod)
        fins = build_finalizers(lambda *x: receiver(GengyList(inner_type, x)), size)
        ident = ctx["_"]
        for i, fin in enumerate(fins):
            nctx = ctx.copy()
            nident = ident + "_" + str(i)
            nctx["_"] = nident
            new_symbol(inner_type, fin, depth, nident, nctx)


class RandomSource(Source):
    def __init__(self, seed: int = 0):
        self.random = random.Random(seed)

    def randint(self, min, max, prod: str = "") -> int:
        return self.random.randint(min, max)

    def random_float(self, min, max, prod: str = "") -> float:
        return self.random.random() * (max - min) + min



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Dict
from typing import Type
from typing import TypeVar

from geneticengine.core.grammar import Grammar
from geneticengine.core.random.sources import Source
from geneticengine.metahandlers.base import MetaHandlerGenerator

min = TypeVar("min", covariant=True)
max = TypeVar("max", covariant=True)


class IntRange(MetaHandlerGenerator):
    """
    IntRange(a,b) restricts ints to be between a and b.
    The range can be dynamically altered before the grammar extraction (Int.__init__.__annotations__["value"] = Annotated[int, IntRange(c,d)].
    """

    def __init__(self, min, max):
        self.min = min
        self.max = max

    def generate(
        self,
        r: Source,
        g: Grammar,
        rec,
        new_symbol,
        depth: int,
        base_type,
        context: dict[str, str],
    ):
        rec(r.randint(self.min, self.max, str(base_type)))

    def __class_getitem__(self, args):
        return IntRange(*args)

    def __repr__(self):
        return f"[{self.min},..,{self.max}]"


class IntList(MetaHandlerGenerator):
    """
    IntList([a_1, .., a_n]) restricts ints to be an element from the list [a_1, .., a_n].
    The range can be dynamically altered before the grammar extraction (Int.__init__.__annotations__["value"] = Annotated[int, IntList[a_1, .., a_n]].
    """

    def __init__(self, elements):
        self.elements = elements

    def generate(
        self,
        r: Source,
        g: Grammar,
        rec,
        new_symbol,
        depth: int,
        base_type,
        context: dict[str, str],
    ):
        rec(r.choice(self.elements, str(base_type)))

    def __class_getitem__(self, args):
        return IntList(*args)

    def __repr__(self):
        return f"[{self.elements}]"


class IntervalRange(MetaHandlerGenerator):
    """
    This metahandler restricts the creation of ranges
    between two integers by forcing a minimum and maximum
    range size, as well as a top limit that the range
    can reach

    This is useful in genomics to generate random windows
    of variable size to scan an input sequence
    """

    def __init__(
        self,
        minimum_length: int,
        maximum_length: int,
        maximum_top_limit: int,
    ):
        """
        :param int minimum_length: Minimum length possible when randomly generating the range
        :param int maximum_length: Maximum length possible when randomly generating the range
        :param int maximum_top_limit: Maximum value the range can reach
        """
        assert maximum_length > minimum_length
        assert maximum_length < maximum_top_limit

        self.minimum_length = minimum_length
        self.maximum_length = maximum_length
        self.maximum_top_limit = maximum_top_limit

    def __class_getitem__(self, args):
        return IntervalRange(*args)

    def generate(
        self,
        r: Source,
        g: Grammar,
        rec,
        new_symbol,
        depth: int,
        base_type,
        context: dict[str, str],
    ):

        range_length = r.randint(self.minimum_length, self.maximum_length)
        start_position = r.randint(0, self.maximum_top_limit - range_length)
        rec((start_position, start_position + range_length))



from __future__ import annotations

from copy import deepcopy
from typing import Callable
from typing import List
from typing import Union

from geneticengine.algorithms.gp.individual import Individual
from geneticengine.core.problems import MultiObjectiveProblem
from geneticengine.core.problems import Problem
from geneticengine.core.problems import SingleObjectiveProblem
from geneticengine.core.random.sources import RandomSource


def create_tournament(
    tournament_size: int,
    problem: Problem,
) -> Callable[[RandomSource, list[Individual], int], list[Individual]]:
    """
    The create_tournament is a function that uses the tournament selection algorithm to select a list of individuals with the best fitness

    Args:
        tournament_size: number of individuals from the population that will be randomly selected
        problem: type of problem that you are trying to solve

    Returns:
        A callable object that returns a list of winners
    """
    assert isinstance(problem, SingleObjectiveProblem)

    def tournament(
        r: RandomSource,
        population: list[Individual],
        n_winners: int,
    ) -> list[Individual]:
        assert isinstance(problem, SingleObjectiveProblem)
        winners = []
        for _ in range(n_winners):
            candidates = [r.choice(population) for _ in range(tournament_size)]
            winner = candidates[0]
            assert winner.fitness is not None
            assert isinstance(winner.fitness, float)
            for o in candidates[1:]:
                assert o.fitness is not None
                assert isinstance(o.fitness, float)
                if (o.fitness > winner.fitness and not problem.minimize) or (
                    o.fitness < winner.fitness and problem.minimize
                ):
                    winner = o
            winners.append(winner)
            candidates.remove(winner)
        return winners

    return tournament


def create_elitism(
    n_elites: int,
) -> Callable[
    [
        list[Individual],
        Problem,
        Callable[[Problem, list[Individual]], Individual],
        Callable[[Individual], float | list[float]],
    ],
    list[Individual],
]:
    """
    The create_elitism is a function that returns the individuals with the best fitness in a generation

    Args:
        n_elites: number of desired  elite Individuals
    Returns:
        A callable object that returns a list of elite Individuals
    """

    def elitism(
        population: list[Individual],
        problem: Problem,
        best_individual_function: Callable[[Problem, list[Individual]], Individual],
        evaluate: Callable[[Individual], float | list[float]],
    ) -> list[Individual]:
        fitnesses = [evaluate(x) for x in population]

        if isinstance(problem, SingleObjectiveProblem):
            assert all(isinstance(x, float) for x in fitnesses)
        else:
            assert all(isinstance(x, list) for x in fitnesses)

        elites: list[Individual] = list()
        population_copy = population.copy()
        i = 0
        while len(elites) < n_elites and i < len(population):

            elite = best_individual_function(problem, population_copy)
            elites.append(elite)
            population_copy.remove(elite)
            i += 1

        return elites

    return elitism


def create_novelties(
    create_individual: Callable[[int], Individual],
    max_depth: int,
) -> Callable[[int], list[Individual]]:
    """
    The create_novelties is a function that returns a list of completely new Individuals

    Args:
        create_individual: callable object that returns a single Individual
        max_depth:
    Returns:
        A callable object that returns a list of new Individuals
    """

    def novelties(n_novelties: int) -> list[Individual]:
        return [create_individual(max_depth) for _ in range(n_novelties)]

    return novelties


def create_lexicase(
    problem: Problem,
) -> Callable[[RandomSource, list[Individual], int], list[Individual]]:
    """
    The create_lexicase is a function that uses the lexicase selection algorithm to select a list of
    Individuals with the best fitness

    Args:
        problem: type of problem that you are trying to solve

    Returns:
        A callable object that returns a list of the selected Individuals
    """
    assert isinstance(problem, MultiObjectiveProblem)

    def lexicase(
        r: RandomSource,
        population: list[Individual],
        n_winners: int,
    ) -> list[Individual]:
        assert isinstance(problem, MultiObjectiveProblem)
        candidates = population.copy()
        assert isinstance(candidates[0].fitness, list)
        n_cases = len(candidates[0].fitness)
        cases = r.shuffle(list(range(n_cases)))
        winners = []

        for _ in range(n_winners):
            candidates_to_check = candidates.copy()

            while len(candidates_to_check) > 1 and len(cases) > 0:
                new_candidates: list[Individual] = list()
                c = cases[0]
                min_max_value = 0
                for i in range(len(candidates_to_check)):
                    checking_candidate = candidates_to_check[i]
                    assert isinstance(checking_candidate.fitness, list)
                    check_value = checking_candidate.fitness[c]
                    if not new_candidates:
                        min_max_value = check_value
                        new_candidates.append(checking_candidate)
                    elif (check_value < min_max_value and problem.minimize[c]) or (
                        check_value > min_max_value and not problem.minimize[c]
                    ):
                        new_candidates.clear()
                        min_max_value = check_value
                        new_candidates.append(checking_candidate)
                    elif check_value == min_max_value:
                        new_candidates.append(checking_candidate)

                candidates_to_check = new_candidates.copy()
                cases.remove(c)

            winner = (
                r.choice(candidates_to_check)
                if len(candidates_to_check) > 1
                else candidates_to_check[0]
            )
            assert isinstance(winner.fitness, list)
            winners.append(winner)
            candidates.remove(winner)
        return winners

    return lexicase



from __future__ import annotations

import time
from typing import Any
from typing import Callable
from typing import NoReturn

import geneticengine.algorithms.gp.generation_steps.cross_over as cross_over
import geneticengine.algorithms.gp.generation_steps.mutation as mutation
import geneticengine.algorithms.gp.generation_steps.selection as selection
from geneticengine.algorithms.callbacks.callback import Callback
from geneticengine.algorithms.callbacks.callback import DebugCallback
from geneticengine.algorithms.callbacks.callback import PrintBestCallback
from geneticengine.algorithms.callbacks.callback import ProgressCallback
from geneticengine.algorithms.callbacks.csv_callback import CSVCallback
from geneticengine.algorithms.gp.individual import Individual
from geneticengine.algorithms.heuristics import Heuristics
from geneticengine.core.grammar import EvolveGrammar
from geneticengine.core.grammar import Grammar
from geneticengine.core.problems import FitnessType
from geneticengine.core.problems import Problem
from geneticengine.core.problems import process_problem
from geneticengine.core.problems import wrap_depth
from geneticengine.core.random.sources import RandomSource
from geneticengine.core.representations.api import Representation
from geneticengine.core.representations.tree.treebased import relabel_nodes_of_trees
from geneticengine.core.representations.tree.treebased import treebased_representation


class GP(Heuristics):
    """Represents the Genetic Programming algorithm. Defaults as given in A
    Field Guide to GP, p.17, by Poli and Mcphee:

    Args:
        grammar (Grammar): The grammar used to guide the search.
        representation (Representation): The individual representation used by the GP program. The default is treebased_representation.
        problem (Problem): The problem we are solving. Either a SingleObjectiveProblem or a MultiObjectiveProblem.
        evaluation_function (Callable[[Any], float]): The fitness function. Should take in any valid individual and return a float. The default is that the higher the fitness, the more applicable is the solution to the problem. Turn on the parameter minimize to switch it around.
        minimize (bool): When switch on, the fitness function is reversed, so that a higher result from the fitness function corresponds to a less fit solution (default = False).
        target_fitness (Optional[float]): Sets a target fitness. When this fitness is reached, the algorithm stops running (default = None).
        favor_less_complex_trees (bool): If set to True, this gives a tiny penalty to more complex (with more nodes) trees to favor simpler trees (default = False).
        randomSource (Callable[[int], RandomSource]): The random source function used by the program. Should take in an integer, representing the seed, and return a RandomSource.
        population_size (int): The population size (default = 200). Apart from the first generation, each generation the population is made up of the elites, novelties, and transformed individuals from the previous generation. Note that population_size > (n_elites + n_novelties + 1) must hold.
        n_elites (int): Number of elites, i.e. the number of best individuals that are preserved every generation (default = 5).
        n_novelties (int): Number of novelties, i.e. the number of newly generated individuals added to the population each generation. (default = 10).
        number_of_generations (int): Number of generations (default = 100).
        max_depth (int): The maximum depth a tree can have (default = 15).
        max_init_depth (int): The maximum depth a tree can have in the initialisation population. Currently only working for tree-based representation and dynamic SGE (default = max_depth).
        min_depth (int): The minimum depth a tree can have (default = None).
        min_init_depth (int): The minimum depth a tree can have in the initialisation population. Only relevant when using the Random_Production initialization method (default = None).
        selection_method (Tuple[str, int]): Allows the user to define the method to choose individuals for the next population (default = ("tournament", 5)).
        ramped_half_and_half (bool): Specify whether you want the population to be initialized ramped half and half. Currently only working for treebased representation (default = True).
        evolve_grammar (EvolveGrammar): You can evolve the grammar throughout the generations by assigning a EvolveGrammar to this parameter (default: None).

        probability_mutation (float): probability that an individual is mutated (default = 0.01).
        probability_crossover (float): probability that an individual is chosen for cross-over (default = 0.9).
        either_mut_or_cro (float | None): Switch evolution style to do either a mutation or a crossover. The given float defines the chance of a mutation. Otherwise a crossover is performed. (default = None),
        hill_climbing (bool): Allows the user to change the standard mutation operations to the hill-climbing mutation operation, in which an individual is mutated to 5 different new individuals, after which the best is chosen to survive (default = False).
        specific_type_mutation (type): Specify a type that is given preference when mutation occurs (default = None),
        specific_type_crossover (type): Specify a type that is given preference when crossover occurs (default = None),
        depth_aware_mut (bool): If chosen, mutations are depth-aware, giving preference to operate on nodes closer to the root. (default = True).
        depth_aware_co (bool): If chosen, crossovers are depth-aware, giving preference to operate on nodes closer to the root. (default = True).

        force_individual (Any): Allows the incorporation of an individual in the first population (default = None).
        seed (int): The seed of the RandomSource (default = 123).

        timer_stop_criteria (bool): If set to True, the algorithm is stopped after the time limit (default = 60 seconds). Then the fittest individual is returned (default = False).
        timer_limit (int): The time limit of the timer.

        save_to_csv (CSVCallback): Saves a CSV file that can be personalized in the CSVCallback class specified (default = None).

        callbacks (List[Callback]): The callbacks to define what is done with the returned prints from the algorithm (default = []).
    """

    # reason for union with noreturn in evaluation function, elitism and elitism: https://stackoverflow.com/questions/51811024/mypy-type-checking-on-callable-thinks-that-member-variable-is-a-method
    grammar: Grammar
    representation: Representation[Any]
    problem: Problem
    evaluation_function: NoReturn | Callable[[Any], float]
    random: RandomSource
    population_size: int
    elitism: (
        NoReturn
        | Callable[
            [
                list[Individual],
                Problem,
                Callable[[Problem, list[Individual]], Individual],
                Callable[[Individual], float | list[float]],
            ],
            list[Individual],
        ]
    )
    mutation: (NoReturn | Callable[[Individual], Individual])
    max_depth: int
    novelty: NoReturn | Callable[[int], list[Individual]]
    final_population: list[Individual]
    callbacks: list[Callback]

    def __init__(
        self,
        grammar: Grammar,
        representation: Representation = treebased_representation,
        # These classes should be replaced by the problem alone.
        problem: Problem | None = None,
        evaluation_function: Callable[
            [Any],
            float,
        ] | None = None,  # DEPRECATE in the next version
        minimize: bool = False,  # DEPRECATE in the next version
        target_fitness: float | None = None,  # DEPRECATE in the next version
        evolve_grammar: EvolveGrammar | None = None,
        favor_less_complex_trees: bool = True,  # DEPRECATE in the next version
        randomSource: Callable[[int], RandomSource] = RandomSource,
        population_size: int = 200,
        n_elites: int = 5,
        n_novelties: int = 10,
        number_of_generations: int = 100,
        max_depth: int = 15,
        max_init_depth: int | None = None,
        min_depth: int | None = None,
        min_init_depth: int | None = None,
        selection_method: tuple[str, int] = ("tournament", 5),
        ramped_half_and_half: bool = False,
        # -----
        # As given in A Field Guide to GP, p.17, by Poli and Mcphee
        probability_mutation: float = 0.01,
        probability_crossover: float = 0.9,
        either_mut_or_cro: float | None = None,
        hill_climbing: bool = False,
        specific_type_mutation: type | None = None,
        specific_type_crossover: type | None = None,
        depth_aware_mut: bool = False,
        depth_aware_co: bool = False,
        # -----
        force_individual: Any = None,
        seed: int = 123,
        # -----
        timer_stop_criteria: bool = False,
        timer_limit: int = 60,
        # -----
        save_to_csv: CSVCallback | None = None,
        # -----
        callbacks: list[Callback] | None = None,
    ):
        assert population_size > (n_elites + n_novelties + 1)

        self.problem: Problem = wrap_depth(
            process_problem(problem, evaluation_function, minimize, target_fitness),
            favor_less_complex_trees,
        )

        self.grammar = grammar
        self.representation = representation
        self.random = randomSource(seed)
        self.seed = seed
        self.population_size = population_size
        self.elitism = selection.create_elitism(n_elites)
        self.max_depth = max_depth
        if max_init_depth:
            assert max_init_depth <= self.max_depth
            self.max_init_depth = max_init_depth
        else:
            self.max_init_depth = self.max_depth
        if min_depth:
            assert min_depth <= self.max_depth
        self.min_depth = min_depth
        if min_init_depth:
            assert min_init_depth <= self.max_init_depth
            self.min_init_depth = min_init_depth
        else:
            self.min_init_depth = self.min_depth
        self.evolve_grammar = evolve_grammar
        self.favor_less_complex_trees = favor_less_complex_trees
        self.novelty = selection.create_novelties(
            self.create_individual,
            max_depth=max_depth,
        )
        self.timer_stop_criteria = timer_stop_criteria
        self.timer_limit = timer_limit
        if callbacks:
            self.callbacks = callbacks
        else:
            self.callbacks = []
        if hill_climbing:
            self.mutation = mutation.create_hill_climbing_mutation(
                self.random,
                self.representation,
                self.grammar,
                max_depth,
                self.keyfitness(),
                5,
                specific_type=specific_type_mutation,
                depth_aware_mut=depth_aware_mut,
            )
        else:
            self.mutation = mutation.create_mutation(
                self.random,
                self.representation,
                self.grammar,
                max_depth,
                specific_type=specific_type_mutation,
                depth_aware_mut=depth_aware_mut,
            )
        self.cross_over = cross_over.create_cross_over(
            self.random,
            self.representation,
            self.grammar,
            max_depth,
            specific_type=specific_type_crossover,
            depth_aware_co=depth_aware_co,
        )
        self.n_novelties = n_novelties
        self.number_of_generations = number_of_generations
        self.probability_mutation = probability_mutation
        self.probability_crossover = probability_crossover
        self.either_mut_or_cro = either_mut_or_cro
        if selection_method[0] == "tournament":
            self.selection = selection.create_tournament(
                selection_method[1],
                self.problem,
            )
        elif selection_method[0] == "lexicase":
            self.selection = selection.create_lexicase(
                self.problem,
            )
        else:
            self.selection = lambda r, ls, n: [x for x in ls[:n]]
        self.force_individual = force_individual
        self.ramped_half_and_half = ramped_half_and_half

        if save_to_csv:
            self.callbacks.append(save_to_csv)

    def evolve(self, verbose=1) -> tuple[Individual, FitnessType, Any]:
        """The main function of the GP object. This function runs the GP
        algorithm over the set number of generations, evolving better
        solutions.

        Args:
            verbose (int): Sets the verbose level of the function (0: no prints, 1: print progress, or 2: print the best individual in each generation).

        Returns a tuple with the following arguments:
            individual (Individual): The fittest individual after the algorithm has finished.
            fitness (float): The fitness of above individual.
            phenotype (Any): The phenotype of the best individual.
        """
        self.callbacks = [
            cb
            for cb in self.callbacks
            if type(cb)
            not in [
                type(DebugCallback()),
                type(PrintBestCallback()),
                type(ProgressCallback()),
            ]
        ]
        if verbose > 2:
            self.callbacks.append(DebugCallback())
        if verbose > 1:
            self.callbacks.append(PrintBestCallback())
        if verbose > 0:
            self.callbacks.append(ProgressCallback())

        # TODO: This is not ramped half and half
        population = self.init_population(self.ramped_half_and_half)
        if self.force_individual is not None:
            population[0] = Individual(
                genotype=relabel_nodes_of_trees(
                    self.force_individual,
                    self.grammar,
                ),
                fitness=None,
            )
        gen = 0
        start = time.time()

        if self.evolve_grammar:
            best_overall = True

        while (not self.timer_stop_criteria and gen < self.number_of_generations) or (
            self.timer_stop_criteria and (time.time() - start) < self.timer_limit
        ):

            npop = self.novelty(self.n_novelties)
            elites = self.elitism(
                population,
                self.problem,
                self.get_best_individual,
                self.evaluate,
            )
            npop.extend(elites)
            spotsLeft = self.population_size - len(npop)
            while spotsLeft > 0:
                if self.either_mut_or_cro:
                    if self.random.random_float(0, 1) < self.either_mut_or_cro:
                        candidate = self.selection(self.random, population, 1)
                        p1 = self.mutation(candidate[0])
                        npop.append(p1)
                        spotsLeft -= 1
                    else:
                        candidates = self.selection(self.random, population, 2)
                        (p1, p2) = self.cross_over(candidates[0], candidates[1])
                        npop.append(p1)
                        npop.append(p2)
                        spotsLeft -= 2
                else:
                    candidates = self.selection(self.random, population, 2)
                    (p1, p2) = candidates[0], candidates[1]
                    if self.random.randint(0, 100) < self.probability_crossover * 100:
                        (p1, p2) = self.cross_over(p1, p2)
                    if self.random.randint(0, 100) < self.probability_mutation * 100:
                        p1 = self.mutation(p1)
                    if self.random.randint(0, 100) < self.probability_mutation * 100:
                        p2 = self.mutation(p2)
                    npop.append(p1)
                    npop.append(p2)
                    spotsLeft -= 2

            population = npop

            best_individual = self.get_best_individual(self.problem, population)

            time_gen = time.time() - start
            for cb in self.callbacks:
                cb.process_iteration(gen + 1, population, time=time_gen, gp=self)

            if self.evolve_grammar:
                if best_overall:
                    inds = elites
                else:
                    for e in elites:
                        npop.remove(e)
                    inds = npop
                best = self.get_best_individual(self.problem, inds)
                prob = best.production_probabilities(
                    self.representation.genotype_to_phenotype,
                    self.grammar,
                )
                self.grammar = self.grammar.update_weights(
                    self.evolve_grammar.learning_rate,
                    prob,
                )
                best_overall = not best_overall
            gen += 1
        self.final_population = population

        for cb in self.callbacks:
            cb.end_evolution()
        return (
            best_individual,
            self.evaluate(best_individual),
            self.representation.genotype_to_phenotype(
                self.grammar,
                best_individual.genotype,
            ),
        )

    def init_population(self, ramped_half_and_half):
        self.representation.method.min_depth = self.min_init_depth
        
        if ramped_half_and_half:
            n_not_ramped = int(self.population_size / 2)
            n_ramped = self.population_size - n_not_ramped 
            pop_ramped = [
                self.create_individual(max((i % self.max_init_depth) + 1, self.grammar.get_min_tree_depth()))
                for i in range(n_ramped)
            ]
            pop_not_ramped = [
                self.create_individual(self.max_init_depth)
                for _ in range(n_not_ramped)
            ]
                
            pop = pop_ramped + pop_not_ramped
        else:
            pop = [
                self.create_individual(self.max_init_depth)
                for _ in range(self.population_size)
            ]
        
        self.representation.depth = self.max_depth
        self.representation.method.min_depth = self.min_depth
        return pop
        



from __future__ import annotations


def abstract(clazz):
    get_gengy(clazz)["abstract"] = True
    return clazz


__builtin_module_name__ = dict().__class__.__module__


def is_builtin(t):
    return t.__module__ == __builtin_module_name__


def get_gengy(t: type) -> dict:
    dic = t.__dict__
    if "__gengy__" not in dic and not is_builtin(t):
        setattr(t, "__gengy__", {})
    return dic.get("__gengy__", {})


def weight(w):
    def weight_w(clazz):
        get_gengy(clazz)["weight"] = w
        return clazz

    return weight_w



from __future__ import annotations

import string
from abc import ABC
from dataclasses import dataclass
from typing import Annotated

from geneticengine.core.decorators import abstract
from geneticengine.metahandlers.vars import VarRange

# Auxiliary "attributes"
an_char = list(string.digits) + list(string.ascii_letters)
s_char = [
    "!",
    "#",
    "$",
    "%",
    "&",
    "(",
    ")",
    "*",
    "+",
    ",",
    "-",
    ".",
    "/",
    ":",
    ";",
    "<",
    "=",
    ">",
    "?",
    "@",
    "[",
    "\\",
    "]",
    "^",
    "_",
    "`",
    "{",
    "}",
    "~",
    "|",
    '"',
    "'",
    " ",
]


# re ::= elementary-re re | elementary-re
@dataclass
class RE(ABC):
    pass


# elementary-re ::= set | range | modifier | char | {match_times} | lookarounds
@abstract
@dataclass
class ElementaryRE(RE):
    pass


# elementary-re ::= [RE] | (RE)
@dataclass
class ElementaryREParens(RE):
    option: Annotated[str, VarRange(["[{}]", "({})"])]
    regex: RE

    def __str__(self):
        return self.option.format(str(self.regex))


# elementary-re ::= \w | \d
@dataclass
class ElementaryREWD(RE):
    option: Annotated[str, VarRange([r"\w", r"\d"])]

    def __str__(self):
        return self.option


# Elementary-RE RE
@dataclass
class ElementaryRERE(RE):
    elementary_regex: ElementaryRE
    regex: RE

    def __str__(self):
        return f"{str(self.elementary_regex)}{self.regex}"


# modifier ::= ^RE | RE. | RE* | RE+ | RE++ | RE? | RE?+ |  RE "|" RE
@dataclass
@abstract
class Modifier(ElementaryRE):
    pass


# modifierSingle ::= ^RE | RE. | RE* | RE+ | RE++ | RE? | RE?+
@dataclass
class ModifierSingle(Modifier):
    modifier: Annotated[str, VarRange(["^", ".", "*", "+", "++", "?", "?+"])]
    regex: RE

    def __str__(self):
        result = str(self.regex)

        if self.modifier == "^":
            result = f"^{result}"
        else:
            result = f"{result}{self.modifier}"

        return result


# modifierOr ::= RE "|" RE
@dataclass
class ModifierOr(Modifier):
    regex1: RE
    regex2: RE

    def __str__(self):
        return f"{self.regex1}|{self.regex2}"


# lookarounds ::= (?<=RE) | (?<!RE) | (?=RE) | (?!RE) | (?:RE) | RE {RE,RE}+
@abstract
@dataclass
class Lookaround(ElementaryRE):
    pass


# lookaroundSingle ::= (?<=RE) | (?<!RE) | (?=RE) | (?!RE) | (?:RE)
@dataclass
class LookaroundSingle(Lookaround):
    lookaround: Annotated[str, VarRange(["?<=", "?<!", "?=", "?!", "?:"])]
    regex: RE

    def __str__(self):
        return f"({self.lookaround}{self.regex})"


# lookaroundComposition ::= RE {RE,RE}+
@dataclass
class LookaroundComposition(Lookaround):
    regex1: RE
    regex2: RE
    regex3: RE

    def __str__(self):
        return f"{self.regex1}{{{self.regex2},{self.regex3}}}+"


# set ::= char | char set
@abstract
@dataclass
class Set(ElementaryRE):
    pass


# char ::= s_char | an_char | an_char | an_char
@dataclass
class Char(Set):
    character: Annotated[str, VarRange(s_char + an_char * 3)]

    def __str__(self):
        return self.character


# char ::= char set
@dataclass
class SetChar(Set):
    character: Char
    _set: Set

    def __str__(self):
        return f"{self.char}{self._set}"


# range ::=  an_char - an_char | an_char - an_char | A-Z | a-z | 0-9
@abstract
@dataclass
class Range(ElementaryRE):
    pass


@dataclass
class RangeAnChar1(Range):
    character1: Annotated[str, VarRange(an_char)]
    character2: Annotated[str, VarRange(an_char)]

    def __str__(self):
        return f"{self.character1}-{self.character2}"


@dataclass
class RangeAnChar2(Range):
    character1: Annotated[str, VarRange(an_char)]
    character2: Annotated[str, VarRange(an_char)]

    def __str__(self):
        return f"{self.character1}-{self.character2}"


@dataclass
class RangeLimits(Range):
    option: Annotated[str, VarRange(["A-Z", "a-z", "0-9"])]

    def __str__(self):
        return self.option


# match_times ::= recur_digit | recur_digit , | recur_digit , recur_digit
@abstract
@dataclass
class MatchTimes(ElementaryRE):
    pass


@abstract
@dataclass
class RecurDigit(MatchTimes):
    pass


@dataclass
class RecurDigitSingle(RecurDigit):
    digit: Annotated[str, VarRange(list(string.digits))]

    def __str__(self):
        return str(self.digit)


@dataclass
class RecurDigitMultiple(RecurDigit):
    digit: Annotated[str, VarRange(list(string.digits))]
    recur_digit: RecurDigit

    def __str__(self):
        return f"{self.digit}{self.recur_digit}"


# match_times ::= recur_digit | recur_digit ,
@dataclass
class MatchTimesSingleRecur(MatchTimes):
    recur_digit: RecurDigit
    option: Annotated[str, VarRange(["", ","])]

    def __str__(self):
        return f"{self.recur_digit}{self.option}"


# match_times ::= recur_digit ,  recur_digit
@dataclass
class MatchTimesDoubleRecur(MatchTimes):
    recur_digit1: RecurDigit
    recur_digit2: RecurDigit

    def __str__(self):
        return f"{self.recur_digit1},{self.recur_digit2}"



from __future__ import annotations

from dataclasses import dataclass
from typing import Any

from geneticengine.core.problems import FitnessType


@dataclass
class Individual:
    genotype: Any
    fitness: FitnessType | None = None

    def __str__(self) -> str:
        return str(self.genotype)

    def count_prods(self, genotype_to_phenotype, g):
        counts = {prod: 0 for prod in g.all_nodes}

        def add_count(ty):
            if ty in counts.keys():
                counts[ty] += 1

        def get_args(no):
            if hasattr(type(no), "__annotations__"):
                return type(no).__annotations__.keys()
            return []

        def counting(node: Any):
            add_count(type(node))
            for base in type(node).__bases__:
                add_count(base)
            for argn in get_args(node):
                counting(getattr(node, argn))

        counting(genotype_to_phenotype(g, self.genotype))
        # self.counts = counts
        return counts

    def production_probabilities(self, genotype_to_phenotype, g):
        counts = self.count_prods(genotype_to_phenotype, g)
        probs = counts.copy()
        for rule in g.alternatives:
            prods = g.alternatives[rule]
            total_counts = 0
            for prod in prods:
                total_counts += counts[prod]
            for prod in prods:
                probs[prod] = counts[prod] / total_counts

        for prod in probs.keys():
            if probs[prod] > 1:
                probs[prod] = 1

        return probs



from __future__ import annotations

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any
from typing import Callable
from typing import Generic
from typing import List
from typing import TypeVar
from typing import Union

from geneticengine.core.decorators import abstract


class Statement(ABC):
    def evaluate(self, **kwargs) -> float:
        return 0

    def evaluate_lines(self, **kwargs) -> Callable[[Any], float]:
        return lambda _: 0


t = TypeVar("t")


@abstract
class Expr(Generic[t]):
    @abstractmethod
    def evaluate(self, **kwargs) -> t:
        ...

    @abstractmethod
    def evaluate_lines(self, **kwargs) -> Callable[[Any], t]:
        ...


@abstract
class Number(Expr[float]):
    pass


@abstract
class Condition(Expr[bool]):
    pass


@abstract
class NumberList(Expr[list[float]]):
    pass


@dataclass
class XAssign(Statement):
    value: Expr

    def evaluate(self, **kwargs) -> float:
        if "x" not in kwargs:
            kwargs["x"] = 1.0
        return self.value.evaluate(**kwargs)

    def evaluate_lines(self, **kwargs) -> Callable[[Any], float]:
        return lambda line: self.value.evaluate_lines(**kwargs)(line)

    def __str__(self):
        return f"x = {self.value}"



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Annotated
from typing import Any
from typing import Callable
from typing import List

from geneticengine.exceptions import GeneticEngineError
from geneticengine.grammars.coding.classes import Number
from geneticengine.grammars.coding.classes import NumberList
from geneticengine.metahandlers.lists import ListSizeBetween
from geneticengine.metahandlers.vars import VarRange


@dataclass
class Max(Number):
    list: NumberList

    def evaluate(self, **kwargs):
        return max(self.list.evaluate(**kwargs))

    def evaluate_lines(self, **kwargs) -> Callable[[Any], float]:
        return lambda line: max(self.list.evaluate_lines(**kwargs)(line))

    def __str__(self) -> str:
        return f"max({self.list})"


@dataclass
class Min(Number):
    list: NumberList

    def evaluate(self, **kwargs):
        return min(self.list.evaluate(**kwargs))

    def evaluate_lines(self, **kwargs) -> Callable[[Any], float]:
        return lambda line: min(self.list.evaluate_lines(**kwargs)(line))

    def __str__(self) -> str:
        return f"min({self.list})"


@dataclass
class Length(Number):
    list: NumberList

    def evaluate(self, **kwargs):
        return len(self.list.evaluate(**kwargs))

    def evaluate_lines(self, **kwargs) -> Callable[[Any], float]:
        return lambda line: len(self.list.evaluate_lines(**kwargs)(line))

    def __str__(self) -> str:
        return f"len({self.list})"


@dataclass
class Combine(NumberList):
    list1: NumberList
    list2: NumberList

    def evaluate(self, **kwargs):
        return self.list1.evaluate(**kwargs) + self.list2.evaluate(**kwargs)

    def evaluate_lines(self, **kwargs) -> Callable[[Any], list[float]]:
        return lambda line: self.list1.evaluate_lines(**kwargs)(
            line,
        ) + self.list2.evaluate_lines(**kwargs)(line)

    def __str__(self) -> str:
        return f"({self.list1} + {self.list2})"


@dataclass
class Literal(NumberList):
    list: Annotated[list[Number], ListSizeBetween(2, 3)]

    def evaluate(self, **kwargs):
        return [v.evaluate(**kwargs) for v in self.list]

    def evaluate_lines(self, **kwargs):
        return lambda line: [v.evaluate_lines(**kwargs)(line) for v in self.list]

    def __str__(self) -> str:
        return str(self.list)


@dataclass
class GetElement(Number):
    list: NumberList
    element: Number

    def evaluate(self, **kwargs):
        list_length = Length(self.list).evaluate(**kwargs)
        return self.list.evaluate(**kwargs)[
            round(self.element.evaluate(**kwargs)) % list_length
        ]

    def evaluate_lines(self, **kwargs):
        def list_length(line):
            return Length(self.list).evaluate_lines(
                **kwargs,
            )(line)

        return lambda line: self.list.evaluate_lines(**kwargs)(line)[
            round(
                self.element.evaluate_lines(
                    **kwargs,
                )(line),
            )
            % list_length(line)
        ]

    def __str__(self) -> str:
        return f"{self.list}[{self.element}]"


@dataclass
class Var(NumberList):
    name: Annotated[str, VarRange(["x", "y", "z"])]

    def evaluate(self, **kwargs):
        return kwargs[self.name]

    def evaluate_lines(self, **kwargs):
        if not hasattr(self, "feature_indices"):
            raise GeneticEngineError(
                "To use geneticengine.grammars.coding.lists.Var.evaluate_lines, one must specify a Var.feature_indices dictionary.",
            )
        return lambda line: line[self.feature_indices[self.name]]

    def __str__(self) -> str:
        return self.name



from __future__ import annotations

from abc import ABC

from geneticengine.core.problems import MultiObjectiveProblem
from geneticengine.core.problems import SingleObjectiveProblem
from geneticengine.core.utils import average_fitness


class Callback(ABC):
    def process_iteration(self, generation: int, population, time: float, gp):
        ...

    def end_evolution(self):
        ...


class DebugCallback(Callback):
    """Example of a callback that prints all the individuals in the population"""

    def process_iteration(self, generation: int, population, time: float, gp):
        for p in population:
            print(f"[{self.__class__}] {p} at generation {generation}")


class ProgressCallback(Callback):
    """Prints the number of the generation"""

    # Currently this only work with GP, doesnt work with Hill Climbing and Random Search
    def process_iteration(self, generation: int, population, time: float, gp):

        best_individual = gp.get_best_individual(gp.problem, population)

        if isinstance(gp.problem, SingleObjectiveProblem):
            fitness = round(gp.evaluate(best_individual), 4)

        elif isinstance(gp.problem, MultiObjectiveProblem):
            fitness = [round(fitness, 4) for fitness in gp.evaluate(best_individual)]
            if len(fitness) > 10:
                fitness = round(average_fitness(best_individual), 4)

        print(
            f"[{self.__class__}] Generation {generation}. Time {time}. Best fitness: {fitness}",
        )


class PrintBestCallback(Callback):
    """Prints the number of the generation"""

    def process_iteration(self, generation: int, population, time: float, gp):
        if isinstance(gp.problem, SingleObjectiveProblem):
            fitness = round(gp.evaluate(population[0]), 4)
        elif isinstance(gp.problem, MultiObjectiveProblem):
            fitness = [round(fitness, 4) for fitness in gp.evaluate(population[0])]
            if len(fitness) > 10:
                fitness = sum(fitness) / len(fitness)

        if not gp.timer_stop_criteria:
            print(
                f"[{self.__class__}] Generation {generation} / {gp.number_of_generations}. Time {time}",
            )
        else:
            print(
                f"[{self.__class__}] Generation {generation}. Time {time} / {gp.timer_stop_criteria}",
            )
        print(f"[{self.__class__}] Best fitness: {fitness}")
        print(f"[{self.__class__}] Best genotype: {population[0].genotype}")



from __future__ import annotations

from typing import Callable
from typing import Optional
from typing import Tuple

from geneticengine.algorithms.gp.individual import Individual
from geneticengine.core.random.sources import Source
from geneticengine.core.representations.api import Representation
from geneticengine.core.representations.tree.treebased import Grammar


def create_cross_over(
    r: Source,
    representation: Representation,
    g: Grammar,
    max_depth: int,
    specific_type: type | None = None,
    depth_aware_co: bool = False,
) -> Callable[[Individual, Individual], tuple[Individual, Individual]]:
    def cross_over_double(
        individual1: Individual,
        individual2: Individual,
    ) -> tuple[Individual, Individual]:
        (g1, g2) = representation.crossover_individuals(
            r,
            g,
            individual1.genotype,
            individual2.genotype,
            max_depth,
            specific_type=specific_type,
            depth_aware_co=depth_aware_co,
        )
        individual1 = Individual(g1)
        individual2 = Individual(g2)
        return (individual1, individual2)

    return cross_over_double



from __future__ import annotations

from typing import Any
from typing import Callable
from typing import Type

import geneticengine.algorithms.gp.generation_steps.mutation as mutation
from geneticengine.algorithms.gp.individual import Individual
from geneticengine.algorithms.heuristics import Heuristics
from geneticengine.core.grammar import Grammar
from geneticengine.core.problems import FitnessType
from geneticengine.core.problems import Problem
from geneticengine.core.problems import process_problem
from geneticengine.core.random.sources import RandomSource
from geneticengine.core.representations.api import Representation
from geneticengine.core.representations.tree.treebased import treebased_representation


class HC(Heuristics):
    """
    Hill Climbing object. Main attribute: evolve

    Args:
        grammar (Grammar): The grammar used to guide the search.
        evaluation_function (Callable[[Any], float]): The fitness function. Should take in any valid individual and return a float. The default is that the higher the fitness, the more applicable is the solution to the problem. Turn on the parameter minimize to switch it around.
        minimize (bool): When switch on, the fitness function is reversed, so that a higher result from the fitness function corresponds to a less fit solution (default = False).
        representation (Representation): The individual representation used by the GP program. The default is treebased_representation.
        randomSource (Callable[[int], RandomSource]): The random source function used by the program. Should take in an integer, representing the seed, and return a RandomSource.
        seed (int): The seed of the RandomSource (default = 123).
        population_size (int): The population size (default = 200). Apart from the first generation, each generation the population is made up of the elites, novelties, and transformed individuals from the previous generation. Note that population_size > (n_elites + n_novelties + 1) must hold.
        number_of_generations (int): Number of generations (default = 100).
        max_depth (int): The maximum depth a tree can have (default = 15).
        force_individual (Any): Allows the incorporation of an individual in the first population (default = None).
    """

    # PROBlEM
    def __init__(
        self,
        g: Grammar,
        evaluation_function: Callable[[Any], float] | None = None,
        representation: Representation = treebased_representation,
        problem: Problem | None = None,
        randomSource: Callable[[int], RandomSource] = RandomSource,
        population_size: int = 200,
        number_of_generations: int = 100,
        max_depth: int = 15,
        minimize: bool = False,
        force_individual: Any = None,
        seed: int = 123,
    ):
        assert population_size >= 1
        self.problem: Problem = process_problem(problem, evaluation_function, minimize)

        self.grammar: Grammar = g
        self.representation = representation
        self.evaluation_function = evaluation_function
        self.random = randomSource(seed)
        self.population_size = population_size
        self.minimize = minimize
        self.mutation = mutation.create_hill_climbing_mutation(
            self.random,
            self.representation,
            self.grammar,
            max_depth,
            self.keyfitness(),
            population_size,
        )
        self.number_of_generations = number_of_generations
        if force_individual is not None:
            self.population = Individual(
                genotype=force_individual,
                fitness=None,
            )
        else:
            self.population = Individual(
                genotype=self.representation.create_individual(
                    self.random,
                    self.grammar,
                    max_depth,
                ),
                fitness=None,
            )

    def evolve(self, verbose=1):
        population = self.population

        for gen in range(self.number_of_generations):
            population = self.mutation(population)
            if verbose == 2:
                print(f"Best population:{population[0]}.")
            if verbose >= 1:
                print(
                    "BEST at",
                    gen + 1,
                    "/",
                    self.number_of_generations,
                    "is",
                    round(self.evaluate(population), 2),
                )
        return (
            population,
            self.evaluate(population),
            self.representation.genotype_to_phenotype(
                self.grammar,
                population.genotype,
            ),
        )



from __future__ import annotations

import time
from typing import Any
from typing import Callable
from typing import Type

from geneticengine.algorithms.callbacks.callback import Callback
from geneticengine.algorithms.callbacks.callback import DebugCallback
from geneticengine.algorithms.callbacks.callback import PrintBestCallback
from geneticengine.algorithms.callbacks.callback import ProgressCallback
from geneticengine.algorithms.callbacks.csv_callback import CSVCallback
from geneticengine.algorithms.gp.individual import Individual
from geneticengine.algorithms.heuristics import Heuristics
from geneticengine.core.grammar import Grammar
from geneticengine.core.problems import FitnessType
from geneticengine.core.problems import Problem
from geneticengine.core.problems import process_problem
from geneticengine.core.problems import wrap_depth
from geneticengine.core.random.sources import RandomSource
from geneticengine.core.representations.api import Representation
from geneticengine.core.representations.tree.treebased import relabel_nodes_of_trees
from geneticengine.core.representations.tree.treebased import treebased_representation


class RandomSearch(Heuristics):
    """
    Random Search object. Main attribute: evolve

    Args:
        grammar (Grammar): The grammar used to guide the search.
        evaluation_function (Callable[[Any], float]): The fitness function. Should take in any valid individual and return a float. The default is that the higher the fitness, the more applicable is the solution to the problem. Turn on the parameter minimize to switch it around.
        minimize (bool): When switch on, the fitness function is reversed, so that a higher result from the fitness function corresponds to a less fit solution (default = False).
        representation (Representation): The individual representation used by the GP program. The default is treebased_representation.
        randomSource (Callable[[int], RandomSource]): The random source function used by the program. Should take in an integer, representing the seed, and return a RandomSource.
        seed (int): The seed of the RandomSource (default = 123).
        population_size (int): The population size (default = 200). Apart from the first generation, each generation the population is made up of the elites, novelties, and transformed individuals from the previous generation. Note that population_size > (n_elites + n_novelties + 1) must hold.
        number_of_generations (int): Number of generations (default = 100).
        max_depth (int): The maximum depth a tree can have (default = 15).
        favor_less_complex_trees (bool): If set to True, this gives a tiny penalty to more complex (with more nodes) trees to favor simpler trees (default = False).
        force_individual (Any): Allows the incorporation of an individual in the first population (default = None).
        save_to_csv (str): Saves a CSV file with the details of all the individuals of all generations.
        save_genotype_as_string (bool): Turn this off if you don't want to safe all the genotypes as strings. This saves memory and a bit of time.
        callbacks (List[Callback]): The callbacks to define what is done with the returned prints from the algorithm (default = []).
    """

    def __init__(
        self,
        grammar: Grammar,
        evaluation_function: Callable[[Any], float] | None = None,
        representation: Representation = treebased_representation,
        problem: Problem | None = None,
        randomSource: Callable[[int], RandomSource] = RandomSource,
        population_size: int = 200,
        number_of_generations: int = 100,
        max_depth: int = 15,
        # now based on depth, maybe on number of nodes?
        favor_less_complex_trees: bool = False,
        minimize: bool = False,
        force_individual: Any = None,
        seed: int = 123,
        save_to_csv: str | None = None,
        save_genotype_as_string: bool = True,
        callbacks: list[Callback] | None = None,
    ):
        assert population_size >= 1

        self.problem: Problem = wrap_depth(
            process_problem(problem, evaluation_function, minimize),
            favor_less_complex_trees,
        )

        self.grammar = grammar
        self.representation = representation
        self.evaluation_function = evaluation_function
        self.random = randomSource(seed)
        self.seed = seed
        self.population_size = population_size
        self.max_depth = max_depth
        self.favor_less_complex_trees = favor_less_complex_trees
        self.minimize = minimize
        self.number_of_generations = number_of_generations
        self.force_individual = force_individual
        if callbacks:
            self.callbacks = callbacks
        else:
            self.callbacks = []
        if save_to_csv:
            c = CSVCallback(
                save_to_csv,
                save_genotype_as_string=save_genotype_as_string,
            )
            self.callbacks.append(c)

    def evolve(self, verbose=1):
        self.callbacks = [
            cb
            for cb in self.callbacks
            if type(cb)
            not in [
                type(DebugCallback()),
                type(PrintBestCallback()),
                type(ProgressCallback()),
            ]
        ]
        if verbose > 2:
            self.callbacks.append(DebugCallback())
        if verbose > 1:
            self.callbacks.append(PrintBestCallback())
        if verbose > 0:
            self.callbacks.append(ProgressCallback())

        best = None
        best_ind = None
        if self.force_individual is not None:
            best_ind = Individual(
                genotype=relabel_nodes_of_trees(
                    self.force_individual,
                    self.grammar,
                ),
                fitness=None,
            )
            best = self.keyfitness()(best_ind)
        for gen in range(self.number_of_generations):
            start = time.time()
            population = [
                self.create_individual(self.max_depth)
                for _ in range(self.population_size)
            ]
            for ind in population:
                f = self.keyfitness()(ind)
                if (best == None) or (f < best):
                    best = f
                    best_ind = ind

            time_gen = time.time() - start
            for cb in self.callbacks:
                cb.process_iteration(gen + 1, population, time=time_gen, gp=self)

        for cb in self.callbacks:
            cb.end_evolution()
        return (
            best_ind,
            self.evaluate(best_ind),
            self.representation.genotype_to_phenotype(
                self.grammar,
                best_ind.genotype,
            ),
        )



from __future__ import annotations

import inspect
import warnings
from abc import ABC
from abc import ABCMeta
from collections import defaultdict
from inspect import isclass
from tracemalloc import start
from typing import Any
from typing import Dict
from typing import Generic
from typing import List
from typing import Set
from typing import Tuple
from typing import Type

from geneticengine.core.decorators import get_gengy
from geneticengine.core.utils import all_init_arguments_typed
from geneticengine.core.utils import get_arguments
from geneticengine.core.utils import get_generic_parameter
from geneticengine.core.utils import get_generic_parameters
from geneticengine.core.utils import has_arguments
from geneticengine.core.utils import is_abstract
from geneticengine.core.utils import is_annotated
from geneticengine.core.utils import is_generic
from geneticengine.core.utils import is_generic_list
from geneticengine.core.utils import is_terminal
from geneticengine.core.utils import strip_annotations


class InvalidGrammarException(Exception):
    """Exception to be raised when a passed node is neither abstract nor
    typed."""

    pass


class EvolveGrammar:
    """Class to be assigned to the evolve_grammar parameter in the GP class.

    Args:
        learning_rate (float): The learning rate with which the grammar is updated each generation (default as in https://arxiv.org/pdf/2103.08389.pdf = 0.01).
    """

    def __init__(self, learning_rate=0.01) -> None:
        self.learning_rate = learning_rate


class Grammar:
    starting_symbol: type
    alternatives: dict[type, list[type]]
    distanceToTerminal: dict[Any, int]
    all_nodes: set[type]
    recursive_prods: set[type]
    terminals: set[
        type
    ]  # todo: both terminals and non_terminals can be obtained by checking if disttoterminal == or!= 0
    non_terminals: set[type]
    abstract_dist_to_t: dict[type, dict[type, int]]

    def __init__(
        self,
        starting_symbol: type,
        considered_subtypes: list[type] | None = None,
        expansion_depthing: bool = False,
    ) -> None:
        self.alternatives: dict[type, list[type]] = {}
        self.starting_symbol = starting_symbol
        self.distanceToTerminal = {int: 0, str: 0, float: 0}
        self.all_nodes = set()
        self.recursive_prods = set()
        self.terminals = set()
        self.non_terminals = set()
        self.abstract_dist_to_t = defaultdict(
            lambda: defaultdict(lambda: 1000000),
        )
        self.considered_subtypes = considered_subtypes or []
        self.expansion_depthing = expansion_depthing

        self.validate()

    def validate(self):
        for c in self.considered_subtypes:
            if is_abstract(c):
                continue
            elif all_init_arguments_typed(c):
                # Raise a warning if there are annotated elements, but no __init__ method.
                d = {x[0]: x[1] for x in inspect.getmembers(c)}
                if (
                    "__annotations__" in d
                    and d["__annotations__"]
                    and d["__init__"].__qualname__ == "object.__init__"
                ):
                    warnings.warn(
                        f"Warning: class {c} looks like it should be a dataclass, but isn't.",
                    )
                else:
                    continue
            else:
                raise InvalidGrammarException(
                    f"Type {c} is not abstract nor has a type-annotated constructor",
                )

    def register_alternative(self, nonterminal: type, nodetype: type):
        """Register a production A->B Call multiple times with same A to
        register many possible alternatives."""
        if not is_abstract(nonterminal):
            raise Exception(
                f"Trying to register an alternative on a non-abstract class: {nonterminal} -> {nodetype}\n"
                f"You may have meant for {nonterminal.__name__} to be abstract. If so, decorate it with @abstract.\n"
                f"(Found in geneticengine.core.decorators)",
            )
        if nonterminal not in self.alternatives:
            self.alternatives[nonterminal] = []
        self.alternatives[nonterminal].append(nodetype)

    def register_type(self, ty: type):
        if ty in self.all_nodes:
            return
        elif is_generic_list(ty):
            gty = get_generic_parameter(ty)
            self.register_type(gty)
            return
        elif is_annotated(ty):
            gty = get_generic_parameter(ty)
            self.register_type(gty)
            return
        elif is_generic(ty):
            for p in get_generic_parameters(ty):
                self.register_type(p)
            return
        self.all_nodes.add(ty)

        parent = ty.mro()[1]
        if parent not in [object, ABC, Generic]:
            assert isinstance(parent, type)
            self.register_type(parent)
            self.register_alternative(parent, ty)

        terminal = False
        if not is_abstract(ty):
            terminal = True
            for (arg, argt) in get_arguments(ty):
                terminal = False
                if isinstance(argt, type) or isinstance(argt, ABCMeta):
                    self.register_type(argt)
                if is_annotated(argt):
                    gen = get_generic_parameter(argt)
                    if is_generic_list(gen):
                        self.register_type(get_generic_parameter(gen))
                    else:
                        self.register_type(gen)

        for st in self.considered_subtypes:
            if issubclass(st, ty):
                self.register_type(st)

        if terminal:
            self.terminals.add(ty)
        else:
            self.non_terminals.add(ty)

    def __repr__(self):
        def wrap(n):
            if hasattr(n, "__name__"):
                return n.__name__
            if hasattr(n, "__metadata__"):
                if is_generic_list(get_generic_parameter(n)):
                    return f"{n.__metadata__[0]} of {wrap(strip_annotations(n))}"
                else:
                    return f"{n.__metadata__[0]}"
            return n

        def format(x):
            def add_weight(prod):
                if "weight" in get_gengy(prod):
                    return f'<{get_gengy(prod)["weight"]}>'
                return ""

            args = ", ".join(
                [f"{a}: {wrap(at)}" for (a, at) in get_arguments(x)],
            )
            return f"{x.__name__}({args}){add_weight(x)}"

        prods = ";".join(
            [
                str(p.__name__)
                + " -> "
                + ("|".join([format(p) for p in self.alternatives[p]]))
                for p in self.alternatives
            ],
        )
        return (
            f"Grammar<Starting={self.starting_symbol.__name__},Productions=[{prods}]>"
        )

    def get_all_symbols(self) -> tuple[set[type], set[type], set[type]]:
        """All symbols in the current grammar, including terminals."""
        keys = {k for k in self.alternatives.keys()}
        sequence = {v for vv in self.alternatives.values() for v in vv}
        return (keys, sequence, sequence.union(keys).union(self.all_nodes))

    def get_distance_to_terminal(self, ty: type) -> int:
        """Returns the current distance to terminal of a given type."""
        if is_annotated(ty):
            ta = get_generic_parameter(ty)
            return self.get_distance_to_terminal(ta)
        elif is_generic_list(ty):
            ta = get_generic_parameter(ty)
            return int(self.expansion_depthing) + self.get_distance_to_terminal(ta)
        elif is_generic(ty):
            return int(self.expansion_depthing) + max(
                self.get_distance_to_terminal(t) for t in get_generic_parameters(ty)
            )
        else:
            return self.distanceToTerminal[ty]

    def get_min_tree_depth(self):
        """Returns the minimum depth a tree must have."""
        return self.distanceToTerminal[self.starting_symbol]

    def get_max_node_depth(self):
        """Returns the maximum minimum depth a node can have."""

        def dist(x):
            return self.distanceToTerminal[x]

        return max(list(map(dist, self.all_nodes)))

    def preprocess(self):
        """Computes distanceToTerminal via a fixpoint algorithm."""
        (keys, _, all_sym) = self.get_all_symbols()
        for s in all_sym:
            self.distanceToTerminal[s] = 1000000
        changed = True

        reachability: dict[type, set[type]] = defaultdict(lambda: set())

        def process_reachability(src: type, dsts: list[type]):
            src = strip_annotations(src)
            ch = False
            src_reach = reachability[src]
            for prod in dsts:
                prod = strip_annotations(prod)
                reach = reachability[prod]
                oldlen = len(reach)
                reach.add(src)
                reach.update(src_reach)
                ch |= len(reach) != oldlen
            return ch

        while changed:
            changed = False
            for sym in all_sym:
                old_val = self.distanceToTerminal[sym]
                val = old_val

                if is_abstract(sym):
                    if sym in self.alternatives:
                        prods = self.alternatives[sym]
                        for prod in prods:
                            val = min(
                                val,
                                int(self.expansion_depthing)
                                + self.distanceToTerminal[prod],
                            )
                            old = self.abstract_dist_to_t[sym][prod]
                            if 1 < old:
                                self.abstract_dist_to_t[sym][prod] = 1
                                changed = True
                            if prod in self.abstract_dist_to_t:
                                for subprod, dist in self.abstract_dist_to_t[
                                    prod
                                ].items():
                                    currdist = self.abstract_dist_to_t[sym][subprod]
                                    candidate = dist + 1
                                    if candidate < currdist:
                                        self.abstract_dist_to_t[sym][
                                            subprod
                                        ] = candidate
                                        changed = True
                        changed |= process_reachability(sym, prods)
                else:
                    if is_terminal(sym, self.non_terminals):
                        if (
                            sym == int or sym == float or sym == str
                        ) and not self.expansion_depthing:
                            val = 0
                        else:
                            val = 1
                    else:
                        args = get_arguments(sym)
                        assert args
                        val = max(
                            1 + self.get_distance_to_terminal(argt)
                            for (_, argt) in args
                        )

                        changed |= process_reachability(
                            sym,
                            [argt for (_, argt) in args],
                        )

                if val < old_val:
                    changed = True
                    self.distanceToTerminal[sym] = val

        for sym in all_sym:
            if sym in reachability[sym]:  # symbol is recursive
                self.recursive_prods.add(sym)
            else:
                pass

    def get_weights(self):
        weights = {prod: get_gengy(prod).get("weight", 1.0) for prod in self.all_nodes}
        return weights

    def update_weights(self, learning_rate, extra_weights):
        weights = self.get_weights()
        for rule in self.alternatives:
            prods = self.alternatives[rule]
            total_weights = 0
            for prod in prods:
                weights[prod] += learning_rate * extra_weights[prod]
                total_weights += weights[prod]
            for prod in prods:
                weights[prod] = weights[prod] / total_weights

        for weight in weights:
            assert weights[weight] >= 0 and weights[weight] <= 1

        starting_symbol = self.starting_symbol
        starting_symbol.__dict__["__gengy__"]["weight"] = weights[starting_symbol]
        nodes = list()
        for node in self.considered_subtypes:
            node.__dict__["__gengy__"]["weight"] = weights[node]
            nodes.append(node)
        self.__init__(starting_symbol, nodes, self.expansion_depthing)
        self.register_type(starting_symbol)
        self.preprocess()
        return self


def extract_grammar(
    considered_subtypes: list[type],
    starting_symbol: type,
    expansion_depthing: bool = False,
):
    """The extract_grammar takes in all the productions of the grammar (nodes)
    and a starting symbol (starting_symbol). It goes through all the nodes and
    constructs a complete grammar that can then be used for search algorithms
    such as Genetic Programming and Hill Climbing.

    Args:
        nodes (list): A list of objects representing tree nodes. Make sure that any node can be produced be the starting symbol.
        starting_symbol (object): The starting symbol of each tree. Makes sure every generated tree by the returned grammar starts with this symbol. Make sure that the starting symbol can produce any object of nodes.

    Returns:
        The grammar
    """
    g = Grammar(starting_symbol, considered_subtypes, expansion_depthing)
    g.register_type(starting_symbol)
    g.preprocess()
    if any(["weight" in get_gengy(p) for p in considered_subtypes]):
        g.update_weights(1, g.get_weights())
    return g



from __future__ import annotations

import re
import string
from abc import ABC
from dataclasses import dataclass
from typing import Annotated

from geneticengine.core.decorators import abstract
from geneticengine.metahandlers.vars import VarRange


# Auxiliary lists of letters
lower_vowel = ["a", "e", "i", "o", "u"]
upper_vowel = ["A", "E", "I", "O", "U"]
lower_consonant = list(re.sub("a|e|i|o|u", "", string.ascii_lowercase))
upper_consonant = list(re.sub("A|E|I|O|U", "", string.ascii_uppercase))


# string :: letter | letter string
@dataclass
class String(ABC):
    pass


# letter ::= char | vowel | consonant
@abstract
@dataclass
class Letter(String):
    pass


# letter string
@dataclass
class LetterString(String):
    letter: Letter
    string: String

    def __str__(self):
        return str(self.letter) + str(self.string)


# char ::= " " | ! | ? | , | .
@dataclass
class Char(Letter):
    value: Annotated[str, VarRange([" ", "!", "?", ",", "."])]

    def __str__(self):
        return self.value


# vowel ::= lower_vowel | upper_vowel
# lower_vowel = a | e | i | o | u
# upper_vowel = A | E | I | O | U
@dataclass
class Vowel(Letter):
    vowel: Annotated[str, VarRange(lower_vowel + upper_vowel)]

    def __str__(self):
        return self.vowel


# consonant ::= lower_consonant | upper_consonant
# lower_consonant ::= b | c | d | f | g | h | j | k | l | ...
# upper_consonant ::= B | C | D | F | G | H | J | K | L | ...
@dataclass
class Consonant(Letter):
    consonant: Annotated[str, VarRange(lower_consonant + upper_consonant)]

    def __str__(self):
        return self.consonant



from __future__ import annotations

from typing import Any
from typing import Callable
from abc import ABC, abstractmethod
import z3
from geneticengine.core.decorators import get_gengy

from geneticengine.core.grammar import Grammar
from geneticengine.core.random.sources import Source
from geneticengine.core.representations.tree.utils import GengyList, relabel_nodes_of_trees
from geneticengine.core.utils import build_finalizers, get_arguments, get_generic_parameter, is_abstract, is_generic_list
from geneticengine.exceptions import GeneticEngineError
from geneticengine.metahandlers.base import is_metahandler


class Initialization_Method(ABC):
    min_depth: int | None = None
    
    @abstractmethod
    def tree_init_method(self, r: Source, g: Grammar, max_depth: int, starting_symbol: type[Any]):
        ...

class Random_Production(Initialization_Method):
    def __init__(
        self, 
        min_depth: int | None = None
    ):
        self.min_depth = min_depth
    
    def tree_init_method(
        self,
        r: Source,
        g: Grammar,
        max_depth: int,
        starting_symbol: type[Any] = int,
    ):
        if self.min_depth:
            assert self.min_depth <= max_depth
            def filter_choices(possible_choices: list[type], depth):
                current_depth = max_depth - depth
                valid_productions = [
                    vp for vp in possible_choices if g.distanceToTerminal[vp] <= depth
                ]
                if (nRecs[0] == 0) and any([  # Are we the last recursive symbol?
                        prod in g.recursive_prods for prod in valid_productions
                    ]  # Are there any  recursive symbols in our expansion?
                ) and current_depth <= self.min_depth:
                    valid_productions = [
                        vp for vp in valid_productions if vp in g.recursive_prods
                    ]  # If so, then only expand into recursive symbols

                return valid_productions
            state = {}

            def final_finalize(x):
                state["final"] = x

            prodqueue = []
            nRecs = [0]

            def handle_symbol(
                next_type,
                next_finalizer,
                depth: int,
                ident: str,
                ctx: dict[str, str],
            ):
                prodqueue.append((next_type, next_finalizer, depth, ident, ctx))
                if next_type in g.recursive_prods:
                    nRecs[0] += 1

            handle_symbol(starting_symbol, final_finalize, max_depth, "root", ctx={})
            
            while prodqueue:
                next_type, next_finalizer, depth, ident, ctx = r.pop_random(prodqueue)
                if next_type in g.recursive_prods:
                    nRecs[0] -= 1
                expand_node(
                    r,
                    g,
                    handle_symbol,
                    filter_choices,
                    next_finalizer,
                    depth,
                    next_type,
                    ident,
                    ctx,
                )
            SMTResolver.resolve_clauses()
            n = state["final"]
            relabel_nodes_of_trees(n, g)
            return n
        else:
            def filter_choices(possible_choices: list[type], depth):
                valid_productions = [
                    vp for vp in possible_choices if g.get_distance_to_terminal(vp) <= depth
                ]
                return valid_productions

            def handle_symbol(
                next_type,
                next_finalizer,
                depth: int,
                ident: str,
                ctx: dict[str, str],
            ):
                expand_node(
                    r,
                    g,
                    handle_symbol,
                    filter_choices,
                    next_finalizer,
                    depth,
                    next_type,
                    ident,
                    ctx,
                )

            state = {}

            def final_finalize(x):
                state["final"] = x

            handle_symbol(starting_symbol, final_finalize, max_depth, "root", ctx={})
            SMTResolver.resolve_clauses()
            n = state["final"]
            relabel_nodes_of_trees(n, g)
            return n

    

class PI_Grow(Initialization_Method):
    def tree_init_method(
        self,
        r: Source,
        g: Grammar,
        max_depth: int,
        starting_symbol: type[Any] = int,
    ):
        state = {}

        def final_finalize(x):
            state["final"] = x

        prodqueue = []
        nRecs = [0]

        def handle_symbol(
            next_type,
            next_finalizer,
            depth: int,
            ident: str,
            ctx: dict[str, str],
        ):
            prodqueue.append((next_type, next_finalizer, depth, ident, ctx))
            if next_type in g.recursive_prods:
                nRecs[0] += 1

        handle_symbol(starting_symbol, final_finalize, max_depth, "root", ctx={})

        def filter_choices(possible_choices: list[type], depth):
            valid_productions = [
                vp for vp in possible_choices if g.distanceToTerminal[vp] <= depth
            ]
            if (nRecs[0] == 0) and any(  # Are we the last recursive symbol?
                [
                    prod in g.recursive_prods for prod in valid_productions
                ],  # Are there any  recursive symbols in our expansion?
            ):
                valid_productions = [
                    vp for vp in valid_productions if vp in g.recursive_prods
                ]  # If so, then only expand into recursive symbols

            return valid_productions

        while prodqueue:
            next_type, next_finalizer, depth, ident, ctx = r.pop_random(prodqueue)
            if next_type in g.recursive_prods:
                nRecs[0] -= 1
            expand_node(
                r,
                g,
                handle_symbol,
                filter_choices,
                next_finalizer,
                depth,
                next_type,
                ident,
                ctx,
            )
        SMTResolver.resolve_clauses()
        n = state["final"]
        relabel_nodes_of_trees(n, g)
        return n

def expand_node(
    r: Source,
    g: Grammar,
    new_symbol,  # Method to find new_symbol (?)
    filter_choices,
    receiver,
    depth,
    starting_symbol: type,
    id: str,
    ctx: dict[str, str],
) -> Any:
    """
    Creates a random node of a given type (starting_symbol)
    """
    if depth < 0:
        raise GeneticEngineError("Recursion Depth reached")
    if depth < g.get_distance_to_terminal(starting_symbol):
        raise GeneticEngineError(
            "There will be no depth sufficient for symbol {} in this grammar (provided: {}, required: {}).".format(
                starting_symbol,
                depth,
                g.get_distance_to_terminal(
                    starting_symbol,
                ),
            ),
        )

    if starting_symbol is int:
        val = r.randint(0, 100, str(starting_symbol))
        SMTResolver.register_const(id, val)
        receiver(val)
        return
    elif starting_symbol is float:
        valf = r.random_float(-100, 100, str(starting_symbol))
        SMTResolver.register_const(id, valf)
        receiver(valf)
        return
    elif starting_symbol is bool:
        valb = r.random_bool(str(starting_symbol))
        SMTResolver.register_const(id, valb)
        receiver(valb)
        return
    elif is_generic_list(starting_symbol):
        ctx = ctx.copy()
        ctx["_"] = id
        r.random_list(
            receiver,
            new_symbol,
            depth,
            starting_symbol,
            ctx,
            str(starting_symbol),
        )
        return
    elif is_metahandler(starting_symbol):
        ctx = ctx.copy()
        ctx["_"] = id
        apply_metahandler(
            r,
            g,
            receiver,
            new_symbol,
            depth,
            starting_symbol,
            ctx,
        )
        return
    else:
        if starting_symbol not in g.all_nodes:
            raise GeneticEngineError(
                f"Symbol {starting_symbol} not in grammar rules.",
            )

        if starting_symbol in g.alternatives:  # Alternatives
            extra_depth = 0
            if is_abstract(starting_symbol) and g.expansion_depthing:
                extra_depth = 1

            compatible_productions = g.alternatives[starting_symbol]
            valid_productions = filter_choices(
                compatible_productions,
                depth - extra_depth,
            )
            if not valid_productions:
                raise GeneticEngineError(
                    "No productions for non-terminal node with type: {} in depth {} (minimum required: {}).".format(
                        starting_symbol,
                        depth - extra_depth,
                        str(
                            [
                                (vp, g.distanceToTerminal[vp])
                                for vp in compatible_productions
                            ],
                        ),
                    ),
                )
            if any(["weight" in get_gengy(p) for p in valid_productions]):
                weights = [get_gengy(p).get("weight", 1.0) for p in valid_productions]
                rule = r.choice_weighted(
                    valid_productions,
                    weights,
                    str(starting_symbol),
                )
            else:
                rule = r.choice(valid_productions, str(starting_symbol))
            new_symbol(rule, receiver, depth - extra_depth, id, ctx)
        else:  # Normal production
            args = get_arguments(starting_symbol)
            ctx = ctx.copy()
            l: list[Any] = []
            for argn, _ in args:
                name = id + "_" + argn
                ctx[argn] = name

                def fn(val, name=name):
                    pass

                l.append(fn)

            fins = build_finalizers(
                mk_save_init(starting_symbol, receiver),
                len(args),
                l,
            )
            for i, (argn, argt) in enumerate(args):
                new_symbol(argt, fins[i], depth - 1, id + "_" + argn, ctx)



def apply_metahandler(
    r: Source,
    g: Grammar,
    receiver,
    new_symbol,
    depth: int,
    ty: type[Any],
    context: dict[str, str],
) -> Any:
    """
    This method applies a metahandler to use a custom generator for things of a given type.
    As an example, AnnotatedType[int, IntRange(3,10)] will use the IntRange.generate(r, recursive_generator)
    The generator is the annotation on the type ("__metadata__").
    """
    metahandler = ty.__metadata__[0]
    base_type = get_generic_parameter(ty)
    return metahandler.generate(
        r,
        g,
        receiver,
        new_symbol,
        depth,
        base_type,
        context,
    )  # todo: last argument


def mk_save_init(starting_symbol: Any, receiver: Callable):
    if isinstance(starting_symbol, type):
        pass
    elif isinstance(starting_symbol, GengyList):
        starting_symbol = starting_symbol.new_like
    else:
        starting_symbol = type(starting_symbol)

    def fin_recv(*x):
        built = starting_symbol(*x)
        built.gengy_init_values = x
        return receiver(built)

    return fin_recv

# TODO: make non static
class SMTResolver:
    clauses: list[Any] = []
    receivers: dict[str, Callable] = {}
    types: dict[str, Callable] = {}

    @staticmethod
    def add_clause(claus, recs: dict[str, Callable]):
        SMTResolver.clauses.extend(claus)
        for k, v in recs.items():
            SMTResolver.receivers[k] = v

    @staticmethod
    def register_type(name, typ):
        SMTResolver.types[name] = SMTResolver.to_z3_typ(typ)

    @staticmethod
    def to_z3_typ(typ):
        return z3.Bool if typ == bool else z3.Int if typ == int else z3.Real

    @staticmethod
    def resolve_clauses():

        if not SMTResolver.receivers:
            return  # don't try to smt solve if we don't need to

        solver = z3.Solver()

        solver.set(":random-seed", 1)
        solver.reset()

        for clause in SMTResolver.clauses:
            solver.add(clause(SMTResolver.types))
        res = solver.check()

        if res != z3.sat:
            raise Exception(f"{solver} failed with {res}")

        model = solver.model()
        for (name, recv) in SMTResolver.receivers.items():
            evaled = model.eval(
                SMTResolver.types[name](
                    name,
                ),
                model_completion=True,
            )

            recv(SMTResolver.get_type(evaled))

        SMTResolver.clauses = []
        SMTResolver.receivers = {}
        SMTResolver.types = {}

    @staticmethod
    def get_type(evaled):
        if type(evaled) == z3.z3.BoolRef:
            evaled = bool(str(evaled))
        elif type(evaled) == z3.z3.IntNumRef:
            evaled = int(str(evaled))
        elif type(evaled) == z3.z3.RatNumRef:
            evaled = eval(str(evaled))
        else:
            raise NotImplementedError(
                f"Don't know what to do with {type(evaled)} {evaled}",
            )
        return evaled

    @staticmethod
    def register_const(ident, val):
        SMTResolver.register_type(ident, type(val))
        ty = SMTResolver.types[ident]
        SMTResolver.clauses.append(lambda _: ty(ident) == val)




from __future__ import annotations

from lark import Lark
from lark import Transformer
from lark import v_args

from geneticengine.metahandlers.smt.lang import *

dsl_grammar = """

expression : "-" expression_plus                             -> minus
           | expression_un                                   -> same

expression_un : expression_bool                              -> same
           | "!" expression_un                               -> nnot
           | expression_bool "==" expression_un              -> binop_eq
           | expression_bool "!=" expression_un              -> binop_neq
           | expression_bool "&&" expression_un              -> binop_and
           | expression_bool _DOUBLEPIPE expression_un       -> binop_or

expression_bool : expression_plus                            -> same
                | expression_plus "<" expression_bool        -> binop_lt
                | expression_plus "<=" expression_bool       -> binop_le
                | expression_plus ">" expression_bool        -> binop_gt
                | expression_plus ">=" expression_bool       -> binop_ge
                | expression_plus "-->" expression_bool      -> binop_imp

expression_plus : expression_fact                           -> same
                | expression_fact "+" expression_plus       -> binop_plus
                | expression_fact "-" expression_plus       -> binop_minus

expression_fact : expression_simple                         -> same
                | expression_simple "*" expression_fact     -> binop_mult
                | expression_simple "/" expression_fact     -> binop_div
                | expression_simple "%" expression_fact     -> binop_mod


expression_simple : "(" expression ")"                      -> same
            | comprehension                                 -> same
            | INTLIT                                        -> int_lit
            | SIGNED_INT                                    -> int_lit
            | FLOATLIT                                      -> float_lit
            | BOOLLIT                                       -> bool_lit
            | VAR ("." VAR)*                                -> var

comprehension : "AllPairs" "(" VAR "," VAR "," VAR ")" "{" expression "}" -> all_pairs

BOOLLIT.5 : "true" | "false"
INTLIT : /[0-9][0-9]*/
FLOATLIT : SIGNED_FLOAT
_DOUBLEPIPE.11 : "||"
VAR : (("a".."z")|"_"|("A".."Z")) (("0".."9")|("a".."z")|("A".."Z"))*

%import common.WS
%import common.CNAME
%import common.SIGNED_INT
%import common.SIGNED_FLOAT

%ignore WS
"""


class TreeToDSL(Transformer):
    def __init__(self):
        pass

    def same(self, args):
        return args[0]

    # Literals

    def var(self, args: list[str]):
        return dVar(args)

    def int_lit(self, args):
        return int(args[0])

    def float_lit(self, args):
        return float(args[0])

    def bool_lit(self, args):
        return str(args[0]) == "true"

    # Expressions

    def minus(self, args):
        raise NotImplementedError()

    def nnot(self, args):
        return dNot(*args)

    def binop_eq(self, args):
        return dEQ(*args)

    def binop_neq(self, args):
        return dNEQ(*args)

    def binop_and(self, args):
        return dAnd(*args)

    def binop_or(self, args):
        return dOr(*args)

    def binop_lt(self, args):
        return dLt(*args)

    def binop_le(self, args):
        return dLE(*args)

    def binop_gt(self, args):
        return dGt(*args)

    def binop_ge(self, args):
        return dGE(*args)

    def binop_imp(self, args):
        return dOr(dNot(args[0]), args[1])

    def binop_plus(self, args):
        return dPlus(*args)

    def binop_minus(self, args):
        raise NotImplementedError()

    def binop_mult(self, args):
        raise NotImplementedError()

    def binop_div(self, args):
        raise NotImplementedError()

    def binop_mod(self, args):
        return dMod(*args)

    def all_pairs(self, args):
        return dAllPairs(args[0], args[1], args[2], args[3])


def mk_parser():
    return Lark(
        dsl_grammar,
        parser="lalr",
        start="expression",
        transformer=TreeToDSL(),
    )


p_expr = mk_parser().parse



from __future__ import annotations

from dataclasses import dataclass
from typing import Annotated
from typing import Any
from typing import Callable

from geneticengine.exceptions import GeneticEngineError
from geneticengine.grammars.coding.classes import Expr
from geneticengine.grammars.coding.classes import Number
from geneticengine.grammars.coding.classes import Statement
from geneticengine.metahandlers.ints import IntRange
from geneticengine.metahandlers.vars import VarRange


@dataclass
class Max(Number):
    left: Number
    right: Number

    def evaluate(self, **kwargs):
        return max(self.left.evaluate(**kwargs), self.right.evaluate(**kwargs))

    def evaluate_lines(self, **kwargs) -> Callable[[Any], float]:
        return lambda line: max(
            self.left.evaluate_lines(**kwargs)(line),
            self.right.evaluate_lines(**kwargs)(line),
        )

    def __str__(self) -> str:
        return f"max({self.left},{self.right})"


@dataclass
class Min(Number):
    left: Number
    right: Number

    def evaluate(self, **kwargs):
        return min(self.left.evaluate(**kwargs), self.right.evaluate(**kwargs))

    def evaluate_lines(self, **kwargs) -> Callable[[Any], float]:
        return lambda line: min(
            self.left.evaluate_lines(**kwargs)(line),
            self.right.evaluate_lines(**kwargs)(line),
        )

    def __str__(self) -> str:
        return f"min({self.left},{self.right})"


@dataclass
class Abs(Number):
    value: Number

    def evaluate(self, **kwargs):
        return abs(self.value.evaluate(**kwargs))

    def evaluate_lines(self, **kwargs) -> Callable[[Any], float]:
        return lambda line: abs(self.value.evaluate_lines(**kwargs)(line))

    def __str__(self) -> str:
        return f"abs({self.value})"


@dataclass
class Plus(Number):
    left: Number
    right: Number

    def evaluate(self, **kwargs):
        return self.left.evaluate(**kwargs) + self.right.evaluate(**kwargs)

    def evaluate_lines(self, **kwargs) -> Callable[[Any], float]:
        return lambda line: self.left.evaluate_lines(**kwargs)(
            line,
        ) + self.right.evaluate_lines(**kwargs)(line)

    def __str__(self) -> str:
        return f"({self.left} + {self.right})"


@dataclass
class Mul(Number):
    left: Number
    right: Number

    def evaluate(self, **kwargs):
        return self.left.evaluate(**kwargs) * self.right.evaluate(**kwargs)

    def evaluate_lines(self, **kwargs) -> Callable[[Any], float]:
        return lambda line: self.left.evaluate_lines(**kwargs)(
            line,
        ) * self.right.evaluate_lines(**kwargs)(line)

    def __str__(self) -> str:
        return f"({self.left} * {self.right})"


@dataclass
class SafeDiv(Number):
    left: Number
    right: Number

    def keep_safe(self, d2):
        if d2 == 0:
            d2 = 0.000001
        return d2

    def evaluate(self, **kwargs):
        d1 = self.left.evaluate(**kwargs)
        d2 = self.right.evaluate(**kwargs)
        return d1 / self.keep_safe(d2)

    def evaluate_lines(self, **kwargs) -> Callable[[Any], float]:
        def d1(line):
            return self.left.evaluate_lines(**kwargs)(line)

        def d2(line):
            return self.keep_safe(
                self.right.evaluate_lines(**kwargs)(line),
            )

        return lambda line: d1(line) / d2(line)

    def __str__(self) -> str:
        return f"({self.left}/{self.right})"


@dataclass
class Literal(Number):
    val: Annotated[int, IntRange(0, 9)]

    def evaluate(self, **kwargs):
        return self.val

    def evaluate_lines(self, **kwargs):
        return lambda _: self.val

    def __str__(self) -> str:
        return str(self.val)


@dataclass
class Var(Number):
    name: Annotated[str, VarRange(["x", "y", "z"])]

    def evaluate(self, **kwargs):
        return kwargs[self.name]

    def evaluate_lines(self, **kwargs):
        if not hasattr(self, "feature_indices"):
            raise GeneticEngineError(
                "To use geneticengine.grammars.coding.expressions.Var.evaluate_lines, one must specify a Var.feature_indices dictionary.",
            )
        return lambda line: line[self.feature_indices[self.name]]

    def __str__(self) -> str:
        return self.name


@dataclass
class XAssign(Statement):
    value: Expr

    def evaluate(self, **kwargs) -> float:
        if "x" not in kwargs:
            kwargs["x"] = 1.0
        return self.value.evaluate(**kwargs)

    def evaluate_lines(self, **kwargs) -> Callable[[Any], float]:
        return lambda line: self.value.evaluate_lines(**kwargs)(line)

    def __str__(self):
        return f"x = {self.value}"


all_operators = [Max, Min, Abs, Plus, Mul, SafeDiv, Literal, Var, XAssign]



from __future__ import annotations

from geneticengine.core.grammar import Grammar
from geneticengine.core.random.sources import RandomSource
from geneticengine.core.representations.tree.treebased import mutate
from geneticengine.core.representations.tree.treebased import random_individual


class RandomMutations:
    def __init__(self, g: Grammar, representation, e, n_generations):
        self.grammar = g
        self.representation = representation
        self.evaluation = e
        self.random = RandomSource()
        self.n_generations = n_generations

    def evolve(self):
        best = 0
        best_ind = None
        i = random_individual(
            self.random,
            self.grammar,
            10,
        )  # Puts in grammar but random_individual takes Processed Grammar
        for _ in range(self.n_generations):
            i = mutate(self.random, self.grammar, i)
            f = self.evaluation(i)
            if f > best:
                best = f
                best_ind = i

        return (best_ind, best)
